import numpy as np


def nan_helper(
    y,
):
    """Helper to handle indices and logical indices of NaNs.

    Parameters:
        - y (np.ndarray): 1d numpy array with possible NaNs

    Returns:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices

    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])
    """

    return np.isnan(y), lambda z: z.nonzero()[0]


def inf_helper(
    y,
):
    """Helper to handle indices and logical indices of infinities.

    Parameters:
        - y (np.ndarray): 1d numpy array with possible infinities

    Returns:
        - infs, logical indices of infinities
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of infinities to 'equivalent' indices

    Example:
        >>> # linear interpolation of infinities
        >>> infs, x= inf_helper(y)
        >>> y[infs]= np.interp(x(infs), x(~infs), y[~infs])
    """

    return np.isinf(y), lambda z: z.nonzero()[0]


def interpolate_nans(
    signal: np.ndarray,
) -> np.ndarray:
    """
    Interpolate NaN values in a given signal.

    Uses linear interpolation to estimate and replace NaN values in the provided
    signal based on the values of non-NaN neighbors.

    Parameters:
        signal (np.ndarray): The input signal array, which might contain NaN values.

    Returns:
        np.ndarray: The signal array with NaN values interpolated.
    """

    nans, x = nan_helper(signal)
    try:
        signal[nans] = np.interp(x(nans), x(~nans), signal[~nans])
    except ValueError:
        # if all values are NaN, we cannot interpolate
        signal = np.full_like(signal, np.nan)
    return signal


def interpolate_infs(
    signal: np.ndarray,
) -> np.ndarray:
    """
    Interpolate infinity values in a given signal.

    Uses linear interpolation to estimate and replace infinity values in the provided
    signal based on the values of non-infinity neighbors.

    Parameters:
        signal (np.ndarray): The input signal array, which might contain infinity values.

    Returns:
        np.ndarray: The signal array with infinity values interpolated.
    """

    infs, x = inf_helper(signal)
    try:
        signal[infs] = np.interp(x(infs), x(~infs), signal[~infs])
    except ValueError:
        # if all values are inf, we cannot interpolate
        signal = np.full_like(signal, np.nan)
    return signal


def interpolate_nans_infs(
    signal: np.ndarray,
) -> np.ndarray:
    """
    Interpolate NaN and infinity values in a given signal.

    Uses linear interpolation to estimate and replace NaN and infinity values in the provided
    signal based on the values of non-NaN and non-infinity neighbors.

    Parameters:
        signal (np.ndarray): The input signal array, which might contain NaN and infinity values.

    Returns:
        np.ndarray: The signal array with NaN and infinity values interpolated.
    """

    signal = interpolate_nans(signal)
    signal = interpolate_infs(signal)
    return signal


# Exception class for the case when a method is not implemented
class NotImplementedError(Exception):
    pass



import sys
from typing import Optional
from typing import Union

import cv2
import numpy as np
import skimage.exposure
import skimage.feature
import skimage.morphology
import torch
from skimage.filters import sobel
from skimage.filters import threshold_otsu
from skimage.filters import threshold_triangle
from skimage.util import img_as_ubyte

from towbintools.deep_learning import augmentation
from towbintools.deep_learning import util
from towbintools.foundation import binary_image
from towbintools.foundation import image_handling
from towbintools.foundation.binary_image import fill_bright_holes


def old_edge_based_segmentation(
    image: np.ndarray,
    pixelsize: float,
    sigma_canny: float = 1,
    low_threshold_ratio: float = 5,
    high_threshold_ratio: float = 2.5,
    kernel_size: int = 5,
    final_threshold_percentile: float = 30,
    **kwargs,
) -> np.ndarray:
    """
    Python adaptation of the OG Matlab code for Sobel-based segmentation

    Parameters:
            image (np.ndarray): The input 2D grayscale image as a NumPy array.
            pixelsize (float): Pixel size to consider when removing small objects.
            sigma_canny (float, optional): Standard deviation for the Gaussian filter used in Canny edge detection. Default is 1.

    Returns:
            np.ndarray: The segmented image as a binary mask (NumPy array).

    Raises:
            ValueError: If the input image is not 2D.
    """

    if image.ndim > 2:
        raise ValueError("Image must be 2D.")

    thresh_otsu = threshold_otsu(image)

    edges = skimage.feature.canny(
        image,
        sigma=sigma_canny,
        low_threshold=thresh_otsu / low_threshold_ratio,
        high_threshold=thresh_otsu / high_threshold_ratio,
    ).astype(np.uint8)

    edges = skimage.morphology.remove_small_objects(
        edges.astype(bool), 3, connectivity=2
    ).astype(np.uint8)

    contours, _ = cv2.findContours(edges, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

    if len(contours) > 10000:
        return np.zeros_like(image).astype(np.uint8)

    edges = binary_image.connect_endpoints(edges, max_distance=200)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (kernel_size, kernel_size))
    edges = (cv2.morphologyEx(edges, cv2.MORPH_CLOSE, kernel) > 0).astype(int)

    mask = fill_bright_holes(image, edges, 10).astype(np.uint8)
    mask = skimage.morphology.remove_small_objects(
        mask.astype(bool), 422.5 / (pixelsize**2), connectivity=2
    ).astype(np.uint8)

    mask = fill_bright_holes(image, mask, 5).astype(np.uint8)

    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)

    if not contours:
        return np.zeros_like(image, dtype=np.uint8)

    out = np.zeros_like(mask)
    cv2.drawContours(out, contours, -1, 1, 1)

    threshold = np.percentile(image[out > 0], final_threshold_percentile)  # type: ignore

    final_mask = image > threshold
    final_mask = skimage.morphology.remove_small_objects(
        final_mask, 422.5 / (pixelsize**2), connectivity=2
    ).astype(np.uint8)

    final_mask = (cv2.morphologyEx(final_mask, cv2.MORPH_CLOSE, kernel) > 0).astype(
        np.uint8
    )

    final_mask = fill_bright_holes(image, final_mask, 10).astype(np.uint8)
    return final_mask


def edge_based_segmentation(
    image: np.ndarray,
    pixelsize: float,
    gaussian_filter_sigma: float = 1,
    kernel_size: int = 30,
    final_threshold_percentile: float = 87.5,
    **kwargs,
) -> np.ndarray:
    """
    Optimized and improved version of the original matlab edge-based segmentation method.

    Parameters:
            image (np.ndarray): The input 2D grayscale image as a NumPy array.
            pixelsize (float): Pixel size to consider when removing small objects.
            gaussian_filter_sigma (float, optional): Standard deviation for the Gaussian filter used in Canny edge detection. Default is 1.

    Returns:
            np.ndarray: The segmented image as a binary mask (NumPy array).

    Raises:
            ValueError: If the input image is not 2D.
    """

    if image.ndim > 2:
        raise ValueError("Image must be 2D.")

    smoothed = skimage.filters.gaussian(image, sigma=gaussian_filter_sigma)

    edge_magnitudes = sobel(smoothed)

    thresh = threshold_otsu(edge_magnitudes)

    edge_mask = (edge_magnitudes > thresh).astype(np.uint8)

    contours, _ = cv2.findContours(edge_mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

    out = np.zeros_like(edge_mask)
    cv2.drawContours(out, contours, -1, 1, 1)

    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (kernel_size, kernel_size))

    out = cv2.morphologyEx(
        out,
        cv2.MORPH_CLOSE,
        kernel,
    )

    new_contours, _ = cv2.findContours(out, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    new_out = np.zeros_like(out)
    cv2.drawContours(new_out, new_contours, -1, 1, 1)

    threshold = np.percentile(image[new_out > 0], final_threshold_percentile)  # type: ignore

    final_mask = image > threshold
    final_mask = skimage.morphology.remove_small_objects(
        final_mask, 422.5 / (pixelsize**2), connectivity=2
    ).astype(np.uint8)

    final_mask = (cv2.morphologyEx(final_mask, cv2.MORPH_CLOSE, kernel) > 0).astype(
        np.uint8
    )

    final_mask = fill_bright_holes(image, final_mask, 10).astype(np.uint8)

    return final_mask


# TODO: Modify this function to allow for non tiled predictions / remove it entirely
def deep_learning_segmentation(
    image,
    model,
    device,
    tiler,
    RGB=True,
    activation=None,
    batch_size=-1,
):
    # Split the image into tiles
    if tiler is None:
        tiles = [image]
    else:
        try:
            tiles = tiler.split(image)
        except Exception:
            print("Error splitting image into tiles. Image shape: ", image.shape)
            sys.exit(1)

    # Prepare tiles for model input
    if RGB:
        tiles = [augmentation.grayscale_to_rgb(tile) for tile in tiles]
    else:
        tiles = [torch.tensor(tile[np.newaxis, ...]).unsqueeze(0) for tile in tiles]

    tiles_batched = (
        util.divide_batch(torch.stack(tiles), batch_size)
        if batch_size > 0
        else [torch.stack(tiles)]
    )

    prediction_tiles = []
    for batch in tiles_batched:
        batch = batch.to(device)

        # Predict
        with torch.no_grad():
            prediction = model(batch)
            if activation == "sigmoid":
                prediction = torch.sigmoid(prediction)
            elif activation == "softmax":
                prediction = torch.softmax(prediction, dim=1)

        # Post-process predictions
        for pred_tile in prediction:
            pred_tile = pred_tile.cpu().numpy()
            pred_tile = np.moveaxis(pred_tile, 0, -1)
            prediction_tiles.append(pred_tile)

    # Merge tiles to get the complete prediction
    if tiler is None:
        pred = prediction_tiles[0]
    else:
        pred = tiler.merge(prediction_tiles)
    mask = (pred > 0.5).astype(np.uint8)

    return mask


def double_threshold_segmentation(image):
    # keep bins 2**8 even though our images are 2**16 because none of the images cover the whole dynamic range of 2**16. This will bin lower abundance signal pixels into fewer histogram bins
    mask = image > _custom_threshold_otsu(image, nbins=2**8)
    mask = cv2.medianBlur(img_as_ubyte(mask), 5)
    mask = skimage.morphology.remove_small_objects(
        mask.astype(bool), 20, connectivity=2
    )
    mask = img_as_ubyte(mask)
    return mask


def segment_image(
    image: Union[str, np.ndarray],
    method: str,
    channels: Optional[list[int]] = None,
    pixelsize: float = None,
    is_zstack=True,
    **kwargs,
) -> np.ndarray:
    """
    Segment an image using the specified method.

    Parameters:
            image (Union[str, np.ndarray]): Input image. If string, it's interpreted as the path to a TIFF file. If ndarray, it's the image data directly.
            method (str): Segmentation method to use. Currently supported: "edge_based", "double_threshold".
            channels (List[int], optional): List of channel indices to keep if reading a multi-channel TIFF file. Default is empty, meaning all channels are kept.
            pixelsize (Optional[float], optional): Physical pixel size to consider for edge-based segmentation. Must be specified if method is "edge_based".
            is_zstack (bool, optional): Whether the input image is a z-stack. Default is True.

            **kwargs: Additional keyword arguments to pass to the segmentation function.
                gaussian_filter_sigma (float, optional): Standard deviation for the Gaussian filter used in edge-based methods. Default is 1.
                final_threshold_percentile (float, optional): Percentile to use for the final thresholding step in edge-based methods. Default is 87.5.
                kernel_size (int, optional): Size of the kernel to use for morphological operations in edge-based methods. Default is 30.


    Returns:
            np.ndarray: The segmented image as a binary mask (NumPy array).

    Raises:
            ValueError: If method is not recognized or if pixelsize is not specified when required.
    """
    if isinstance(image, str):
        image = image_handling.read_tiff_file(image, channels_to_keep=channels)

    if method == "edge_based":
        if pixelsize is None:
            raise ValueError("Pixelsize must be specified for edge-based segmentation.")

        def segment_fn(x):
            return edge_based_segmentation(x, pixelsize, **kwargs)

    elif method == "double_threshold":
        segment_fn = double_threshold_segmentation

    else:
        raise ValueError("Invalid segmentation method.")

    if is_zstack:
        mask = np.zeros(
            (image.shape[0], image.shape[-2], image.shape[-1]), dtype=np.uint8
        )
        for i, plane in enumerate(image):
            mask[i] = segment_fn(plane).squeeze()
        return mask

    return segment_fn(image)


def _mode_limited_mean(image, nbins=2**8, histogram=None):
    """
    Brocher, Jan. (2014). Qualitative and Quantitative Evaluation of Two New Histogram Limiting Binarization Algorithms.
    International Journal of Image Processing (IJIP). 8. 30-48.
    Section 2.2 - Mode Limited Mean (MoLiM)
    """
    # basic idea of the paper and MoLiM is that most thresholding algorithms work on histograms
    # if most of the image is 'background', then the histogram will be dominated by a single background peak, which will affect how the threshold is calculated
    # MoLiM is the mean of the image with some of the background already classified as background
    # this limited mean value can either itself be a threshold, or can be used as a starting point for more sophisticated thresholding algorithms
    if histogram is None:
        hist, bin_centers = skimage.exposure.histogram(image, nbins=nbins)
    else:
        hist, bin_centers = histogram

    mode = bin_centers[hist.argmax()]
    molim = np.mean(image[image > mode])
    return molim


def _custom_threshold_otsu(image, nbins=2**8):
    # for consistent behaviour with regards to dtype and nbins, explicitly convert image to float
    # this is because skimage.filters.threshold_... and skimage.exposure.histogram ignore nbins in case of integer dtype
    # additionally, I think skimage trims histogram values below image.min() and above image.max() since they are zero
    # so all in all, when trying to troubleshoot threshold algorithm behaviour, and make it consistent across different channels, rescale the image as float in range (0, 1)

    # save the image range and dtype to later convert threshold back to what is expected
    image_range = image.min(), image.max()
    image_dtype = image.dtype
    image = skimage.exposure.rescale_intensity(image, out_range=(0, 1))
    # calculate histogram once and reuse in later thresholding steps
    histogram = skimage.exposure.histogram(image, nbins=nbins)
    hist_vals, bin_centers = histogram
    # limited_mean lies somewhere in the histogram peak for background
    limited_mean = _mode_limited_mean(image, histogram=histogram)
    # threshold_triangle by-and-large finds the threshold for the bottom of background peak
    # so it's a good starting point for otsu threshold, which can struggle in the case when background is so large as a proportion of the image that the overall histogram appears to be unimodal
    thresh_lower_bound = image.min()
    i = 0
    while thresh_lower_bound < limited_mean:
        # >= to allow possibility of thresh_lower_bound == limited_mean
        thresh_lower_bound = threshold_triangle(
            image[image >= thresh_lower_bound], nbins=nbins
        )
        if i > 64:
            break
        i += 1

    # don't start right at lower bound because the histogram information below thresh_lower_bound may be useful in determining a good threshold in the last iteration
    # for example, when thresh_lower_bound is already a decent threshold, in which case the the very information used to calculate it would have been discarded
    thresh = image.min()
    i = 0
    while thresh < thresh_lower_bound:
        # crop histogram to only include values above the current threshold
        hist_vals = hist_vals[bin_centers > thresh]
        bin_centers = bin_centers[bin_centers > thresh]
        thresh = threshold_otsu(hist=(hist_vals, bin_centers))

        if i > 64:
            break
        i += 1

    # rescale threshold back to original image range and dtype since threshold is in (0, 1)
    # rather than do the math ourselves, let skimage do it as it also handles float quantization
    thresh = skimage.exposure.rescale_intensity(
        np.array(thresh).reshape(1, 1), in_range=(0, 1), out_range=image_range
    ).astype(image_dtype)[0, 0]

    return thresh



import numpy as np
from scipy.signal import find_peaks


def heatmap_to_keypoints_1D(heatmap, height_threshold=0.5):
    """
    Convert a heatmap to keypoints by finding the peaks.
    """
    peaks = []
    for i in range(heatmap.shape[0]):
        try:
            peaks_i, peaks_i_dict = find_peaks(heatmap[i], height=height_threshold)
            # keep only the highest peak
            best_peak = np.argmax(peaks_i_dict["peak_heights"])
            peaks.append(peaks_i[best_peak])
        except ValueError:
            peaks.append(np.nan)
    peaks = np.array(peaks)
    return peaks



from typing import Optional

import cv2
import numpy as np
from scipy.integrate import simpson
from scipy.interpolate import interp1d
from scipy.interpolate import splev
from scipy.interpolate import splprep
from scipy.ndimage import binary_fill_holes
from scipy.signal import savgol_filter
from scipy.stats import kurtosis
from scipy.stats import skew
from skimage.feature import graycomatrix
from skimage.feature import graycoprops
from skimage.measure import regionprops
from skimage.measure import regionprops_table
from skimage.measure import shannon_entropy
from skimage.util import img_as_ubyte

from towbintools.foundation import binary_image
from towbintools.foundation.image_quality import LAPM
from towbintools.foundation.image_quality import LAPV
from towbintools.foundation.image_quality import normalized_variance_measure
from towbintools.foundation.image_quality import TENG
from towbintools.foundation.image_quality import TENG_VARIANCE
from towbintools.straightening import Warper


AVAILABLE_MASK_FEATURES = [
    "length",
    "volume",
    "area",
    "width_mean",
    "width_median",
    "width_std",
    "width_cv",
    "width_skew",
    "width_kurtosis",
    "width_max",
    "width_middle",
    "bending_energy",
]

AVAILABLE_IMAGE_FEATURES = [
    "tenegrad",
    "tenegrad_variance",
    "laplacian_variance",
    "modified_laplacian",
    "normalized_variance",
]

FEATURES_TO_COMPUTE_AT_MOLT = [
    "volume",
    "length",
    "area",
    "fluo",
    "width",
]


def get_available_regionprops():
    from skimage.measure._regionprops import (
        PROP_VALS,
        _require_intensity_image,
    )

    mask_props = [prop for prop in PROP_VALS if prop not in _require_intensity_image]
    image_props = [prop for prop in PROP_VALS if prop in _require_intensity_image]

    return mask_props, image_props


SKIMAGE_MASK_FEATURES, SKIMAGE_IMAGE_FEATURES = get_available_regionprops()


def get_available_mask_features():
    """
    Get a list of available worm features.

    Returns:
        list: The list of available worm features.
    """
    return AVAILABLE_MASK_FEATURES


def get_features_to_compute_at_molt():
    """
    Get a list of features to compute at molt.

    Returns:
        list: The list of features to compute at molt.
    """
    return FEATURES_TO_COMPUTE_AT_MOLT


def compute_mask_volume(
    straightened_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the volume of a straightened mask using its radius and pixel size.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for volume calculation.

    Returns:
        float: The computed volume of the worm.
    """

    worm_radius = np.sum(straightened_mask, axis=0) / 2
    return np.sum(np.pi * (worm_radius**2)) * (pixelsize**3)


def compute_mask_area(
    worm_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the area of a worm mask.

    Parameters:
        worm_mask (np.ndarray): The worm mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for area calculation.

    Returns:
        float: The computed are of the worm.
    """

    return np.sum(worm_mask) * (pixelsize**2)


def compute_mask_length(
    straightened_mask: np.ndarray,
    pixelsize: float,
) -> float:
    """
    Compute the length of a straightened mask using the sum of its pixels along the 0-axis.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for length calculation.

    Returns:
        float: The computed length of the worm.
    """
    return np.sum(np.sum(straightened_mask, axis=0) > 0) * pixelsize


def compute_mask_average_width(
    straightened_mask: np.ndarray,
    pixelsize: float,
    aggregation: str = "mean",
) -> float:
    """
    Compute the average width of a straightened mask.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for width calculation.
        aggregation (str): The aggregation method to use to compute the width.
                           Can be either 'mean' or 'median'. (default: 'mean')

    Returns:
        float: The computed average width of the worm.
    """

    worm_widths = np.sum(straightened_mask, axis=0) * pixelsize
    worm_widths = worm_widths[worm_widths > 0]

    if aggregation == "mean":
        return np.mean(worm_widths)
    elif aggregation == "median":
        return np.median(worm_widths)
    else:
        raise ValueError('Aggregation must be one of "mean" or "median".')


def compute_width_profile(
    straightened_mask: np.ndarray,
    pixelsize: float,
    smooth: bool = True,
    savgol_window: int = 21,
    savgol_order: int = 3,
) -> np.ndarray:
    """
    Compute the width profile of a straightened mask.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for width calculation.
        smooth (bool): Whether to apply a Savitzky-Golay filter to the width profile. (default: True)
        savgol_window (int): The window size of the Savitzky-Golay filter. (default: 21)
        savgol_order (int): The order of the Savitzky-Golay filter. (default: 3)

    Returns:
        np.ndarray: The computed and smoothed width profile. If smoothing fails, the unsmoothed profile is returned.
    """

    profile = np.sum(straightened_mask, axis=0) * pixelsize

    profile = profile[profile > 0]

    if not smooth:
        return profile

    try:
        return savgol_filter(profile, savgol_window, savgol_order)
    except ValueError:
        return profile


def compute_max_width(
    width_profile: np.ndarray,
    window_size: int = 10,
) -> float:
    """
    Compute the maximum width of a worm from its width profile by taking the maximum value and averaging the values around it.

    Parameters:
        width_profile (np.ndarray): The width profile of the worm.
        window_size (int): The size of the window to average around the maximum. (default: 10)

    Returns:
        float: The maximum width of the worm.
    """

    try:
        max_width_index = np.argmax(width_profile)
    except ValueError:
        return np.nan
    return np.mean(
        width_profile[
            max_width_index - window_size : max_width_index + window_size + 1
        ]  # noqa : E203
    )


def compute_mid_width(
    width_profile: np.ndarray,
    window_size: int = 10,
):
    """
    Compute the width of a worm at its midpoint from its width profile by averaging the values around the midpoint.

    Parameters:
        width_profile (np.ndarray): The width profile of the worm.
        window_size (int): The size of the window to average around the midpoint. (default: 10)

    Returns:
        float: The width of the worm at its midpoint.
    """

    mid_width_index = len(width_profile) // 2
    return np.mean(
        width_profile[
            mid_width_index - window_size : mid_width_index + window_size + 1
        ]  # noqa : E203
    )


def compute_mask_morphological_features(
    straightened_mask: np.ndarray,
    pixelsize: float,
    features: list[str],
) -> dict:
    """
    Compute a set of morphological features for a straightened mask.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for various calculations.
        features (list): The list of features to compute.

    Returns:
        dict: A dictionary containing the computed features.
    """

    # if any of the feature contains "width", compute the width profile
    if any("width" in feature for feature in features):
        width_profile = compute_width_profile(straightened_mask, pixelsize=pixelsize)

    feature_dict = {}
    for feature in features:
        if feature == "length":
            feature_dict["length"] = compute_mask_length(
                straightened_mask, pixelsize=pixelsize
            )
        elif feature == "volume":
            feature_dict["volume"] = compute_mask_volume(
                straightened_mask, pixelsize=pixelsize
            )
        elif feature == "area":
            feature_dict["area"] = compute_mask_area(
                straightened_mask, pixelsize=pixelsize
            )
        elif feature == "width_mean":
            feature_dict["width_mean"] = np.mean(width_profile)
        elif feature == "width_median":
            feature_dict["width_median"] = np.median(width_profile)
        elif feature == "width_std":
            feature_dict["width_std"] = np.std(width_profile)
        elif feature == "width_cv":
            feature_dict["width_cv"] = np.std(width_profile) / np.mean(width_profile)
        elif feature == "width_skew":
            feature_dict["width_skew"] = skew(width_profile)
        elif feature == "width_kurtosis":
            feature_dict["width_kurtosis"] = kurtosis(width_profile)
        elif feature == "width_max":
            feature_dict["width_max"] = compute_max_width(width_profile)
        elif feature == "width_middle":
            feature_dict["width_middle"] = compute_mid_width(width_profile)
        elif feature == "bending_energy":
            feature_dict["bending_energy"] = compute_bending_energy_mask(
                straightened_mask, pixelsize
            )
        else:
            raise ValueError(f"Feature {feature} not recognized.")

    # set all features to NaN if they are 0
    for key in feature_dict:
        if feature_dict[key] == 0:
            feature_dict[key] = np.nan

    return feature_dict


def compute_image_features(
    image: np.ndarray,
    features: list,
) -> dict:
    """
    Compute a set of image features for an image.

    Parameters:
        image (np.ndarray): The input image as a NumPy array.
        features (list): The list of features to compute.

    Returns:
        dict: A dictionary containing the computed features.
    """

    feature_dict = {}
    for feature in features:
        if feature == "tenegrad":
            feature_dict["tenegrad"] = TENG(image)
        elif feature == "tenegrad_variance":
            feature_dict["tenegrad_variance"] = TENG_VARIANCE(image)
        elif feature == "laplacian_variance":
            feature_dict["laplacian_variance"] = LAPV(image)
        elif feature == "modified_laplacian":
            feature_dict["modified_laplacian"] = LAPM(image)
        elif feature == "normalized_variance":
            feature_dict["normalized_variance"] = normalized_variance_measure(image)
        else:
            raise ValueError(f"Feature {feature} not recognized.")

    return feature_dict


def compute_mask_type_features(
    straightened_mask: np.ndarray,
    pixelsize: float,
) -> list:
    """
    Compute a series of morphological features for a straightened mask including
    length, volume, volume per length, width measures, entropy, and region properties.

    Parameters:
        straightened_mask (np.ndarray): The straightened mask as a NumPy array.
        pixelsize (float): The size of a pixel, used for various calculations.

    Returns:
        list: A list containing the computed features in the following order:
              [worm_length, worm_volume, volume_per_length, width_mean, width_std,
              width_cv, entropy_mask, eccentricity, solidity, permimeter]
    """
    # compute worm length
    worm_length = compute_mask_length(straightened_mask, pixelsize=pixelsize)

    # compute worm volume
    worm_volume = compute_mask_volume(straightened_mask, pixelsize=pixelsize)

    # compute worm width
    worm_widths = np.sum(straightened_mask, axis=0) * pixelsize
    worm_widths = worm_widths[worm_widths > 0]

    width_std = np.std(worm_widths)
    width_mean = np.mean(worm_widths)
    width_cv = width_std / width_mean

    volume_per_length = worm_volume / worm_length

    # compute entropy
    entropy_mask = shannon_entropy(straightened_mask)

    try:
        other_properties = regionprops(straightened_mask.astype(np.uint8))[0]
        eccentricity = other_properties.eccentricity
        solidity = other_properties.solidity
        permimeter = other_properties.perimeter
    except IndexError:
        eccentricity = 0
        solidity = 0
        permimeter = 0

    return [
        worm_length,
        worm_volume,
        volume_per_length,
        width_mean,
        width_std,
        width_cv,
        entropy_mask,
        eccentricity,
        solidity,
        permimeter,
    ]


# Features for nuclei classification


def intensity_std(
    regionmask: np.ndarray,
    intensity_image: np.ndarray,
) -> float:
    """
    Compute the standard deviation of the intensity values within a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.

    Returns:
        float: The standard deviation of the intensity values within the region.
    """
    return np.std(intensity_image[regionmask])


def intensity_skew(
    regionmask: np.ndarray,
    intensity_image: np.ndarray,
) -> float:
    """
    Compute the skewness of the intensity values within a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.

    Returns:
        float: The skewness of the intensity values within the region."""

    return skew(intensity_image[regionmask])


def intensity_kurtosis(
    regionmask: np.ndarray,
    intensity_image: np.ndarray,
) -> float:
    """
    Compute the kurtosis of the intensity values within a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.

    Returns:
        float: The kurtosis of the intensity values within the region."""

    return kurtosis(intensity_image[regionmask])


def compute_haralick_features(
    image: np.ndarray,
) -> list:
    """
    Compute Haralick texture features for an image.

    Parameters:
        image (np.ndarray): The intensity image.

    Returns:
        list: A list of Haralick texture features.
    """

    # Calculate the Grey-Level Co-Occurrence Matrix
    glcm = graycomatrix(
        img_as_ubyte(image),
        distances=[1],
        angles=[0],
        levels=256,
        symmetric=True,
        normed=True,
    )
    # Calculate properties
    contrast = graycoprops(glcm, "contrast")[0, 0]
    dissimilarity = graycoprops(glcm, "dissimilarity")[0, 0]
    homogeneity = graycoprops(glcm, "homogeneity")[0, 0]
    energy = graycoprops(glcm, "energy")[0, 0]
    correlation = graycoprops(glcm, "correlation")[0, 0]

    return [contrast, dissimilarity, homogeneity, energy, correlation]


def compute_patch_features(
    regionmask: np.ndarray,
    intensity_image: np.ndarray,
    patch_size: int = 64,
) -> list:
    """
    Compute a set of features for a patch around a region.

    Parameters:
        regionmask (np.ndarray): The mask of the region.
        intensity_image (np.ndarray): The intensity image.
        patch_size (int): The size of the patch. (default: 64)

    Returns:
        list: A list of features for the patch.
    """

    # Get the centroid of the region
    centroid = regionprops(regionmask.astype("uint8"))[0].centroid
    # Create a patch of the defined size around the centroid
    minr = int(centroid[0] - patch_size / 2)
    maxr = int(centroid[0] + patch_size / 2)
    minc = int(centroid[1] - patch_size / 2)
    maxc = int(centroid[1] + patch_size / 2)

    if minr < 0:
        minr = 0
        maxr = patch_size
    if minc < 0:
        minc = 0
        maxc = patch_size
    if maxr >= intensity_image.shape[0]:
        maxr = intensity_image.shape[0] - 1
        minr = maxr - patch_size
    if maxc >= intensity_image.shape[1]:
        maxc = intensity_image.shape[1] - 1
        minc = maxc - patch_size

    patch = intensity_image[minr:maxr, minc:maxc]

    patch_basic_intensity_features = [
        np.max(patch),
        np.min(patch),
        np.mean(patch),
        np.std(patch),
        skew(patch.ravel()),
        kurtosis(patch.ravel()),
    ]
    patch_texture_features = compute_haralick_features(patch)
    patch_advanced_intensity_features = [shannon_entropy(patch)]

    patch_features = (
        patch_basic_intensity_features
        + patch_texture_features
        + patch_advanced_intensity_features
    )

    return patch_features


def compute_base_label_features(
    mask_of_label: np.ndarray,
    intensity_image: np.ndarray,
    features: list[str],
    extra_properties: list[str],
) -> list:
    """
    Compute a set of features for a single label.

    Parameters:
        mask_of_label (np.ndarray): The mask of the label.
        intensity_image (np.ndarray): The intensity image.
        features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.

    Returns:
        list: A list of features for the label.
    """

    properties = regionprops_table(
        mask_of_label,
        intensity_image=intensity_image,
        properties=features,
        extra_properties=extra_properties,
    )
    feature_vector = []
    for feature in properties:
        feature_vector.append(properties[feature][0])
    return feature_vector


def get_context(
    current_label: int,
    mask_of_current_label: np.ndarray,
    mask_of_labels: np.ndarray,
    num_closest: int = 5,
) -> np.ndarray:
    """
    Returns the mask of the closest labels to the current label in a label image.

    Parameters:
        current_label (int): The label of the current region.
        mask_of_current_label (np.ndarray): The mask of the current region.
        mask_of_labels (np.ndarray): The mask of all regions.
        num_closest (int): The number of closest regions to return. (default: 5)

    Returns:
        np.ndarray: The mask of the closest regions.
    """

    mask_of_all_other_labels = mask_of_labels.copy()
    mask_of_all_other_labels[mask_of_all_other_labels == current_label] = 0
    mask_of_all_other_labels = mask_of_all_other_labels.astype("uint8")

    if num_closest == -1:
        return mask_of_all_other_labels
    else:
        centroid_current_label = regionprops(mask_of_current_label)[0].centroid
        centroid_other_labels = regionprops(mask_of_all_other_labels)

        # find the num_closest labels
        closest_labels = sorted(centroid_other_labels, key=lambda x: np.linalg.norm(np.array(x.centroid) - np.array(centroid_current_label)))[:num_closest]  # type: ignore
        closest_labels = [x.label for x in closest_labels]
        binary_mask_of_closest_labels = np.isin(mask_of_labels, closest_labels).astype(
            "uint8"
        )
        mask_of_closest_labels = mask_of_labels.copy()
        mask_of_closest_labels[binary_mask_of_closest_labels == 0] = 0
        return mask_of_closest_labels.astype("uint8")


def get_context_features(
    mask_of_labels: np.ndarray,
    intensity_image: np.ndarray,
    features: list[str],
    extra_properties: list[str],
) -> list:
    """
    Compute a set of features for the context of a label.

    Parameters:
        mask_of_labels (np.ndarray): The mask of all regions.
        intensity_image (np.ndarray): The intensity image.
        features (list): The list of features to compute.
        extra_properties (list): The list of extra properties to compute.

    Returns:
        list: A list of aggregated features for the context of the label (mean and std of each desired feature).
    """

    properties = regionprops_table(
        mask_of_labels,
        intensity_image=intensity_image,
        properties=features,
        extra_properties=extra_properties,
    )
    context_feature_vector = []
    for feature in properties:
        context_feature_vector.append(np.mean(properties[feature]))
        context_feature_vector.append(np.std(properties[feature]))
    return context_feature_vector


def compute_bending_energy(
    midline_points: np.array,
    widths: np.array,
    E: float = 1.0,
    smooth: Optional[float] = None,
) -> float:
    """
    Compute bending energy of a midline considering variable width.

    Parameters:
        midline_points (np.array): array of shape (n, 2) containing centerline x,y coordinates
        widths (np.array): array of shape (n,) containing width at each point
        E (float): Young's modulus (set to 1.0 for relative comparisons)
        smooth (float): smoothing factor for spline fitting. If None, the default scipy smoothing is applied. (default: None)

    Returns:
        float: bending energy
    """

    # Fit splines to both centerline and width
    tck, u = splprep([midline_points[:, 0], midline_points[:, 1]], s=smooth)

    # Create width interpolation function
    width_interp = interp1d(np.linspace(0, 1, len(widths)), widths, kind="cubic")

    # Generate points along the spline for analysis
    u_new = np.linspace(0, 1, 1000)
    x_new, y_new = splev(u_new, tck)
    widths_new = width_interp(u_new)

    # Compute derivatives for curvature
    dx_du, dy_du = splev(u_new, tck, der=1)
    dx2_du2, dy2_du2 = splev(u_new, tck, der=2)

    # Compute curvature
    numerator = dx_du * dy2_du2 - dy_du * dx2_du2
    denominator = (dx_du * dx_du + dy_du * dy_du) ** 1.5
    curvature = numerator / denominator

    # Compute differential arc length
    ds = np.sqrt(dx_du * dx_du + dy_du * dy_du) * (u_new[1] - u_new[0])

    # Compute second moment of area (I) assuming circular cross-section
    # I = (π/64) * d⁴ for a circular cross-section
    I = np.pi * (widths_new**4) / 64.0  # noqa: E741

    # Compute local flexural rigidity
    EI = E * I

    # Compute bending energy = ∫ (EI/2) κ²ds
    local_energy = (EI / 2) * curvature**2 * ds

    total_energy = simpson(local_energy, x=u_new)

    return total_energy


def compute_bending_energy_mask(
    mask: np.ndarray,
    pixelsize: float,
    E: float = 1.0,
    smooth: Optional[float] = None,
    savgol_window: int = 21,
    savgol_order: int = 3,
) -> float:
    """
    Compute bending energy of a binary mask by extracting its midline and width profile.

    Parameters:
        mask (np.ndarray): Binary mask.
        pixelsize (float): Physical size of a pixel, used for width calculation.
        E (float): Young's modulus (set to 1.0 for relative comparisons)
        smooth (float): smoothing factor for spline fitting. If None, the default scipy smoothing is applied. (default: None)
        savgol_window (int): window size for Savitzky-Golay filter. (default: 21)
        savgol_order (int): order of Savitzky-Golay filter. (default: 3)

    Returns:
        float: bending energy
    """

    try:
        # Extract midline and width profile
        mask = binary_fill_holes(mask)
        mask_for_midline = cv2.medianBlur(mask.astype(np.uint8), 5)
        mask = binary_image.get_biggest_object(mask.astype(np.uint8))
        mask_for_midline = binary_image.get_biggest_object(
            mask_for_midline.astype(np.uint8)
        )
        warper = Warper.from_img(mask, mask_for_midline)
        midline = warper.splines[0]
        length = warper.length
        midline = midline(np.linspace(0, length, 1000))

        straightened_mask = warper.warp_2D_img(
            mask,
            0,
            interpolation_order=0,
            preserve_range=True,
            preserve_dtype=True,
        )

        widths = compute_width_profile(
            straightened_mask, pixelsize, savgol_window, savgol_order
        )

        bending_energy = compute_bending_energy(midline, widths)
    except Exception as e:
        print(e)
        bending_energy = np.nan
    return bending_energy



import cv2
import numpy as np


def normalized_variance_measure(
    image: np.ndarray,
) -> float:
    """
    Compute the normalized variance measure of an image.

    Parameters:
            image (np.ndarray): The input image as a numpy array.

    Returns:
            float: The computed normalized variance value.
    """
    mean = np.mean(image)
    var = np.var(image)
    return np.divide(var, mean)


# This code is from : https://github.com/vismantic-ohtuprojekti/qualipy/blob/master/qualipy/utils/focus_measure.py

# Python implementations for focus measure operators described
# in "Analysis of focus measure operators for shape-from-focus"
# (Pattern recognition, 2012) by Pertuz et al.

# LICENSE

# The MIT License (MIT)

# Copyright (c) 2015 QualiPy developers

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


def LAPV(img):
    """Implements the Variance of Laplacian (LAP4) focus measure
    operator. Measures the amount of edges present in the image.

    :param img: the image the measure is applied to
    :type img: np.ndarray
    :returns: np.float32 -- the degree of focus
    """
    return np.std(cv2.Laplacian(img, cv2.CV_64F)) ** 2  # type: ignore


def LAPM(img):
    """Implements the Modified Laplacian (LAP2) focus measure
    operator. Measures the amount of edges present in the image.

    :param img: the image the measure is applied to
    :type img: np.ndarray
    :returns: np.float32 -- the degree of focus
    """
    kernel = np.array([-1, 2, -1])
    laplacianX = np.abs(cv2.filter2D(img, -1, kernel))
    laplacianY = np.abs(cv2.filter2D(img, -1, kernel.T))
    return np.mean(laplacianX + laplacianY)


def TENG(img):
    """Implements the Tenengrad (TENG) focus measure operator.
    Based on the gradient of the image.

    :param img: the image the measure is applied to
    :type img: np.ndarray
    :returns: np.float32 -- the degree of focus
    """
    gaussianX = cv2.Sobel(img, cv2.CV_64F, 1, 0)  # type: ignore
    gaussianY = cv2.Sobel(img, cv2.CV_64F, 1, 0)  # type: ignore
    return np.mean(gaussianX * gaussianX + gaussianY * gaussianY)


def MLOG(img):
    """Implements the MLOG focus measure algorithm.

    :param img: the image the measure is applied to
    :type img: np.ndarray
    :returns: np.float32 -- the degree of focus
    """
    return np.max(cv2.convertScaleAbs(cv2.Laplacian(img, cv2.CV_64F)))  # type: ignore


def TENG_VARIANCE(img):
    """Implements the Tenengrad Variance focus measure operator.

    :param img: the image the measure is applied to
    :type img: np.ndarray
    :returns: np.float32 -- the degree of focus
    """

    gaussianX = cv2.Sobel(img, cv2.CV_64F, 1, 0)  # type: ignore
    gaussianY = cv2.Sobel(img, cv2.CV_64F, 1, 0)  # type: ignore

    G = np.sqrt(gaussianX**2 + gaussianY**2)
    return np.var(G)



from .deep_learning_tools import create_pretrained_segmentation_model
from .deep_learning_tools import create_segmentation_model
from .deep_learning_tools import load_segmentation_model_from_checkpoint
from .utils import augmentation
from .utils import util

__all__ = [
    "create_pretrained_segmentation_model",
    "create_segmentation_model",
    "load_segmentation_model_from_checkpoint",
    "load_keypoint_detection_model_from_checkpoint",
    "create_keypoint_detection_model",
    "augmentation",
    "util",
]



import torch.nn as nn


class FocalTverskyLoss(nn.Module):
    def __init__(self, weight=None, size_average=True, ignore_index=-1):
        super().__init__()
        self.ignore_index = ignore_index

    def forward(self, inputs, targets, smooth=100, alpha=0.3, beta=0.7, gamma=4 / 3):
        # comment out if your model contains a sigmoid or equivalent activation layer
        # inputs = F.sigmoid(inputs)

        # flatten label and prediction tensors
        inputs = inputs.view(-1)
        targets = targets.view(-1)

        # remove elements with ignore_index from the loss calculation
        mask = targets != self.ignore_index
        inputs = inputs[mask]
        targets = targets[mask]

        # True Positives, False Positives & False Negatives
        TP = (inputs * targets).sum()
        FP = ((1 - targets) * inputs).sum()
        FN = (targets * (1 - inputs)).sum()

        Tversky = (TP + smooth) / (TP + alpha * FP + beta * FN + smooth)
        FocalTversky = (1 - Tversky + 1e-10) ** gamma

        return FocalTversky


class BCELossWithIgnore(nn.Module):
    def __init__(self, ignore_index=-1):
        super().__init__()
        self.bce_loss = nn.BCELoss(reduction="none")
        self.ignore_index = ignore_index

    def forward(self, input, target):
        # Calculate BCE loss

        # Create a mask for non-ignored values
        mask = (target != self.ignore_index).float()

        # Set the ignored values to 0
        target = target * mask

        # Calculate the loss
        loss = self.bce_loss(input, target)

        # Apply the mask to the loss
        masked_loss = loss * mask

        # Return the mean of the masked loss
        return masked_loss.sum() / mask.sum().clamp(min=1e-8)


class PeakWeightedMSELoss(nn.Module):
    def __init__(self, ignore_index=-1, peak_weight=3.0):
        super().__init__()
        self.ignore_index = ignore_index
        self.peak_weight = peak_weight

    def forward(self, input, target):
        weights = 1.0 + self.peak_weight * target  # Higher weight for higher values
        return (weights * (input - target) ** 2).mean()



import os
from math import ceil
from math import floor

import torch

# import torch.nn as nn
# from efficientnet_pytorch.utils import Conv2dStaticSamePadding


def divide_batch(batch, n):
    for i in range(0, batch.shape[0], n):
        yield batch[i : i + n, ::]  # noqa: E203


def get_closest_upper_multiple(dim, multiple):
    return int(multiple * ceil(dim / multiple))


def get_closest_lower_multiple(dim, multiple):
    return int(multiple * floor(dim / multiple))


def adjust_tensor_dimensions(source_tensor, target_tensor_shape):
    # Squeeze out unnecessary dimensions from source tensor
    adjusted_tensor = source_tensor.squeeze()
    # Add necessary dimensions to match the target tensor shape
    for dim, size in enumerate(target_tensor_shape):
        if dim >= len(adjusted_tensor.shape):
            adjusted_tensor = adjusted_tensor.unsqueeze(dim)
    return adjusted_tensor


def rename_keys_and_adjust_dimensions(model, pretrained_model):
    """Util function to easily load a pretrained model's weights into a model with the same architecture but different module names, etc."""
    assert len(model.state_dict().keys()) == len(
        pretrained_model.keys()
    ), f"The number of keys in the model ({len(model.state_dict().keys())}) and the pretrained model ({len(pretrained_model.keys())}) do not match."
    new_state_dict = {}
    for key, old_key in zip(model.state_dict().keys(), pretrained_model.keys()):
        target_tensor = model.state_dict()[key]
        source_tensor = pretrained_model[old_key]
        # Adjust the dimensions of the source tensor to match the target tensor
        if source_tensor.shape != target_tensor.shape:
            source_tensor = adjust_tensor_dimensions(source_tensor, target_tensor.shape)
        new_state_dict[key] = source_tensor
    return new_state_dict


# def change_first_conv_layer_input(model, new_in_channels):
#     for name, module in model.named_children():
#         if isinstance(module, Conv2dStaticSamePadding):
#             # Extract parameters from the original layer
#             out_channels = module.out_channels
#             kernel_size = module.kernel_size
#             stride = module.stride
#             dilation = module.dilation
#             groups = module.groups
#             bias = module.bias is not None

#             # Create a new Conv2d layer with the desired number of input channels
#             new_conv = Conv2dStaticSamePadding(
#                 new_in_channels,
#                 out_channels,
#                 kernel_size=kernel_size,  # type: ignore
#                 stride=stride,
#                 dilation=dilation,  # type: ignore
#                 groups=groups,
#                 bias=bias,
#                 image_size=512,
#             )  # I think the image_size parameter can be anything but is required anyway

#             # Replace the original layer with the new one
#             setattr(model, name, new_conv)
#             break
#         if isinstance(module, nn.Conv2d):
#             # Extract parameters from the original layer
#             out_channels = module.out_channels
#             kernel_size = module.kernel_size
#             stride = module.stride
#             padding = module.padding
#             dilation = module.dilation
#             groups = module.groups
#             bias = module.bias is not None

#             # Create a new Conv2d layer with the desired number of input channels
#             new_conv = nn.Conv2d(
#                 new_in_channels,
#                 out_channels,
#                 kernel_size=kernel_size,  # type: ignore
#                 stride=stride,
#                 padding=padding,
#                 dilation=dilation,  # type: ignore
#                 groups=groups,
#                 bias=bias,
#             )

#             # Replace the original layer with the new one
#             setattr(model, name, new_conv)
#             break

#         elif len(list(module.children())) > 0:
#             # Recursively call the function for nested modules (e.g., nn.Sequential)
#             change_first_conv_layer_input(module, new_in_channels)
#             break  # Break after modifying the first conv layer in any nested module


# def change_last_fc_layer_output(model, new_out_features):
#     # Reverse iterate through the model's children
#     for name, module in reversed(list(model.named_children())):
#         if isinstance(module, nn.Linear):
#             # Extract parameters from the original layer
#             in_features = module.in_features

#             # Create a new Linear layer with the desired output features
#             new_linear = nn.Linear(in_features, new_out_features)

#             # Replace the original layer with the new one
#             setattr(model, name, new_linear)
#             break
#         elif len(list(module.children())) > 0:
#             # Recursively call the function for nested modules (e.g., nn.Sequential)
#             change_last_fc_layer_output(module, new_out_features)
#             break  # Break after modifying the last linear layer in any nested module


def get_input_channels_from_checkpoint(checkpoint_path):
    checkpoint = torch.load(checkpoint_path, map_location="cpu")
    state_dict = checkpoint["state_dict"]

    input_channels = 0
    for key in state_dict.keys():
        if "conv" in key and "weight" in key:
            input_channels = state_dict[key].shape[1]
            break
    return input_channels


def create_lightweight_checkpoint(input_path, output_path):
    """
    Load existing PyTorch Lightning checkpoint and save a lightweight version
    """
    # Load the full checkpoint
    print(f"Loading checkpoint from: {input_path}")
    checkpoint = torch.load(input_path, map_location="cpu")

    # Display original file size
    original_size = os.path.getsize(input_path) / (1024 * 1024)  # MB
    print(f"Original file size: {original_size:.2f} MB")

    # Create lightweight checkpoint with only essential information
    lightweight_checkpoint = {
        "state_dict": checkpoint["state_dict"],
        "hyper_parameters": checkpoint.get("hyper_parameters", {}),
    }

    # Optionally keep other useful metadata
    optional_keys = ["epoch", "global_step", "pytorch-lightning_version"]
    for key in optional_keys:
        if key in checkpoint:
            lightweight_checkpoint[key] = checkpoint[key]

    # Save the lightweight version
    torch.save(lightweight_checkpoint, output_path)

    # Display new file size and savings
    new_size = os.path.getsize(output_path) / (1024 * 1024)  # MB
    savings = ((original_size - new_size) / original_size) * 100

    print(f"Lightweight file size: {new_size:.2f} MB")
    print(f"Space saved: {savings:.1f}%")

    # Show what was removed
    removed_keys = set(checkpoint.keys()) - set(lightweight_checkpoint.keys())
    if removed_keys:
        print(f"Removed keys: {removed_keys}")

    return lightweight_checkpoint



import numpy as np
from scipy import interpolate
from scipy.interpolate import interp1d
from scipy.signal import medfilt
from whittaker_eilers import WhittakerSmoother

from towbintools.foundation.utils import interpolate_nans


def resize_series_to_length(series, new_length, kind="linear"):
    """
    Resize a 1D or 2D series to a new length using interpolation.

    Parameters:
        series (np.ndarray): The time series to resize.
        new_length (int): The desired length of the resized series.
        kind (str, optional): The type of interpolation to use. (default: "linear")

    Returns:
        np.ndarray: The resized series of the specified length.

    Raises:
        ValueError: If the input series is not 1D or 2D.
    """
    series = np.asarray(series)
    length = series.shape[-1]
    if length == new_length:
        return series
    x_old = np.linspace(0, 1, length)
    x_new = np.linspace(0, 1, new_length)
    if series.ndim == 1:
        f = interp1d(x_old, series, kind=kind, fill_value="extrapolate")
        return f(x_new)
    elif series.ndim == 2:
        f = interp1d(x_old, series, axis=1, kind=kind, fill_value="extrapolate")
        return f(x_new)
    else:
        raise ValueError("Input series must be 1D or 2D array.")


def pad_series_to_length(series, new_length, pad_value=0):
    """
    Pad a 1D or 2D series to a new length with a specified pad value.

    Parameters:
        series (np.ndarray): The time series to pad.
        new_length (int): The desired length of the padded series.
        pad_value (int, optional): The value to use for padding. Default is 0

    Returns:
        np.ndarray: The padded series of the specified length.

    Raises:
        ValueError: If the input series is not 1D or 2D.
    """
    series = np.asarray(series)
    length = series.shape[-1]
    if length == new_length:
        return series
    if series.ndim == 1:
        return np.pad(
            series, (0, new_length - length), mode="constant", constant_values=pad_value
        )
    elif series.ndim == 2:
        return np.pad(
            series,
            ((0, 0), (0, new_length - length)),
            mode="constant",
            constant_values=pad_value,
        )
    else:
        raise ValueError("Input series must be 1D or 2D array.")


def crop_series_to_length(series, new_length):
    """
    Crop a 1D or 2D series to a new length.

    Parameters:
        series (np.ndarray): The time series to crop.
        new_length (int): The desired length of the cropped series.

    Returns:
        np.ndarray: The cropped series of the specified length.

    Raises:
        ValueError: If the input series is not 1D or 2D.
    """
    series = np.asarray(series)
    length = series.shape[-1]
    if length == new_length:
        return series
    if series.ndim == 1:
        return series[0:new_length]
    elif series.ndim == 2:
        return series[:, 0:new_length]
    else:
        raise ValueError("Input series must be 1D or 2D array.")


def random_crop_series_to_length(series, new_length):
    """
    Randomly crop a 1D or 2D series to a new length.

    Parameters:
        series (np.ndarray): The time series to crop.
        new_length (int): The desired length of the cropped series.

    Returns:
        np.ndarray: The cropped series of the specified length.

    Raises:
        ValueError: If the input series is not 1D or 2D.
    """
    series = np.asarray(series)
    length = series.shape[-1]

    if length == new_length:
        return series
    if series.ndim == 1:
        start = np.random.randint(0, length - new_length + 1)
        return series[start : start + new_length]
    elif series.ndim == 2:
        start = np.random.randint(0, length - new_length + 1)
        return series[:, start : start + new_length]
    else:
        raise ValueError("Input series must be 1D or 2D array.")


def correct_series_with_classification(series, worm_type):
    """
    Remove the points of non-worms from the time series and interpolate them back.

    Parameters:
        series (np.ndarray): The time series of values.
        worm_type (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.

    Returns:
        np.ndarray: The corrected time series.
    """

    # Set the series of non worms to NaN
    non_worms_indices = worm_type != "worm"
    series_worms = series.copy()
    series_worms[non_worms_indices] = np.nan

    try:
        # Interpolate the NaNs
        series_worms = interpolate_nans(series_worms)
    except ValueError:
        return np.full(series.shape, np.nan)

    return series_worms


def filter_series_with_classification(series, worm_type):
    """Remove the points of non-worms from the time series.

    Parameters:
        series (np.ndarray): The time series of values.
        worm_type (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.

    Returns:
        np.ndarray: The filtered time series."""

    # Set the series of non worms to NaN
    non_worms_indices = worm_type != "worm"
    series_worms = series.copy()
    series_worms[non_worms_indices] = np.nan

    return series_worms


def interpolate_larval_stage(series, time, ecdysis, larval_stage, n_points=100):
    if larval_stage < 1 or larval_stage > 4:
        raise ValueError("The larval stage must be between 1 and 4.")

    previous_ecdys = ecdysis[larval_stage - 1]
    ecdys = ecdysis[larval_stage]

    # check that the molts are correct
    if (np.isnan(previous_ecdys) or np.isnan(ecdys)) or previous_ecdys > ecdys:
        return np.full(n_points, np.nan), np.full(n_points, np.nan)

    # convert ecdysis times to int to get the index
    previous_ecdys = int(previous_ecdys)
    ecdys = int(ecdys)

    interpolated_time = np.linspace(time[previous_ecdys], time[ecdys], n_points)
    interpolated_series = interpolate.interp1d(time, series, kind="linear")(
        interpolated_time
    )

    return interpolated_time, interpolated_series


def interpolate_entire_development(series, time, ecdysis, n_points=100):
    interpolated_time = np.full((4, n_points), np.nan)
    interpolated_series = np.full((4, n_points), np.nan)
    for larval_stage in range(1, 5):
        (
            interpolated_time_stage,
            interpolated_series_stage,
        ) = interpolate_larval_stage(time, series, ecdysis, larval_stage, n_points)

        interpolated_time[larval_stage - 1, :] = interpolated_time_stage
        interpolated_series[larval_stage - 1, :] = interpolated_series_stage

    return interpolated_time, interpolated_series


def interpolate_entire_development_classified(
    series, time, ecdysis, worm_type, n_points=100
):
    time = filter_series_with_classification(time, worm_type)
    series = filter_series_with_classification(series, worm_type)

    return interpolate_entire_development(time, series, ecdysis, n_points)


def compute_exponential_series_at_time_classified(
    series: np.ndarray,
    time: np.ndarray,
    worm_types: np.ndarray,
    fit_width: int = 10,
) -> float:
    """
    Compute the value of a time series at a given time(s) using linear regression on a logarithmic transformation of the data.

    This function uses linear regression on a logarithmic transformation of the volume data to predict
    the volume at the specified hatch time and end-molts. Only data points where `worm_types` is "worm"
    are used for fitting. The function returns the volume at desired time.

    Parameters:
        volume (np.ndarray): A time series representing volume.
        time (np.ndarray): The time(s) at which the volume is to be computed.
        worm_types (np.ndarray): An array indicating the type of each entry in the volume time series. Expected values are "worm", "egg", etc.
        fit_width (int, optional): Width for the linear regression fit used in computing the volume. (default: 10)

    Returns:
        np.ndarray: Volume at desired time(s).
    """

    def compute_single_time(time: float) -> float:
        if np.isfinite(time):
            fit_x = np.arange(
                max(0, int(time - fit_width)),
                min(len(series), int(time + fit_width)),
                dtype=int,
            )

            filtered_fit_x = fit_x[np.where(worm_types[fit_x] == "worm")]
            if filtered_fit_x.size != 0:
                fit_y = np.log(series[filtered_fit_x])
                try:
                    p = np.polyfit(filtered_fit_x, fit_y, 1)
                except Exception as e:
                    print(
                        f"Caught an exception while interpolating volume at time {time}, returning nan : {e}"
                    )
                    return np.nan
                series_at_time = np.exp(np.polyval(p, time))
            else:
                series_at_time = np.nan
        else:
            series_at_time = np.nan

        return float(series_at_time)

    result = np.array([compute_single_time(t) for t in time])

    return result


def smooth_series_classified(
    series: np.ndarray,
    series_time,
    worm_types: np.ndarray,
    lmbda=0.0075,
    order=2,
    medfilt_window=5,
) -> np.ndarray:
    """
    Compute the series at the given time points using the worm types to classify the points. The series is first corrected for incorrect segmentation, then median filtered to remove outliers, then smoothed using Whittaker-Eilers smoothing.

    Parameters:
        series (np.ndarray): The time series.
        series_time (np.ndarray): The time points of the original series. If None, the time points are assumed to be the indices of the series.
        worm_types (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.
        lmbda (float, optional): The smoothing parameter for the Whittaker-Eilers smoothing. Default provides good results for our volume curves when series_time is in hours. (default: 0.0075)
        order (int, optional): The order of the penalty of the Whittaker-Eilers smoother. (default: 2)
        medfilt_window (int, optional): The window size for the median filter. (default: 5)
    Returns:
        np.ndarray: The series at the given time(s).
    """

    # Check if the series has any non nan values
    if np.all(np.isnan(series)):
        return np.full(series.shape, np.nan)

    # Interpolate the nans
    series = interpolate_nans(series)

    # Remove the points of non-worms from the time series and interpolate them back
    series = correct_series_with_classification(series, worm_types)

    smoothed_series = _smooth_series(
        series,
        series_time,
        lmbda=lmbda,
        order=order,
        medfilt_window=medfilt_window,
    )
    return smoothed_series


def smooth_series(
    series: np.ndarray,
    series_time,
    lmbda=0.0075,
    order=2,
    medfilt_window=5,
) -> np.ndarray:
    """
    Compute the series at the given time points using the worm types to classify the points. The series is first corrected for incorrect segmentation, then median filtered to remove outliers, then smoothed using Whittaker-Eilers smoothing.

    Parameters:
        series (np.ndarray): The time series.
        series_time (np.ndarray): The time points of the original series. If None, the time points are assumed to be the indices of the series.
        lmbda (float, optional): The smoothing parameter for the Whittaker-Eilers smoothing. Default provides good results for our volume curves when series_time is in hours. (default: 0.0075)
        order (int, optional): The order of the penalty of the Whittaker-Eilers smoother. (default: 2)
        medfilt_window (int, optional): The window size for the median filter. (default: 5)
    Returns:
        np.ndarray: The series at the given time(s).
    """

    # Check if the series has any non nan values
    if np.all(np.isnan(series)):
        return np.full(series.shape, np.nan)

    # Interpolate the nans
    series = interpolate_nans(series)

    smoothed_series = _smooth_series(
        series,
        series_time,
        lmbda=lmbda,
        order=order,
        medfilt_window=medfilt_window,
    )
    return smoothed_series


def _smooth_series(
    series: np.ndarray,
    series_time: np.ndarray,
    lmbda: float = 0.0075,
    order: int = 2,
    medfilt_window: int = 5,
) -> np.ndarray:
    # Median filter to remove outliers
    series = medfilt(series, medfilt_window)

    if series_time is not None:
        nan_series_time = np.isnan(series_time)
        non_nan_series = series[~nan_series_time]
        non_nan_series_time = series_time[~nan_series_time]
        whittaker_smoother = WhittakerSmoother(
            lmbda=lmbda,
            order=order,
            data_length=len(non_nan_series),
            x_input=non_nan_series_time,
        )

        smoothed_series = whittaker_smoother.smooth(non_nan_series)
    else:
        whittaker_smoother = WhittakerSmoother(
            lmbda=lmbda, order=order, data_length=len(series)
        )

        smoothed_series = whittaker_smoother.smooth(series)

    smoothed_series = np.array(smoothed_series)

    # pad the smoothed series to the original shape
    smoothed_series = np.pad(
        smoothed_series,
        (0, series.shape[0] - smoothed_series.shape[0]),
        mode="constant",
        constant_values=np.nan,
    )

    # Interpolate the nans again
    smoothed_series = interpolate_nans(smoothed_series)

    return smoothed_series


def compute_series_at_time_classified(
    series: np.ndarray,
    time: np.ndarray,
    series_time: np.ndarray,
    worm_types: np.ndarray,
    lmbda=0.0075,
    medfilt_window=5,
    bspline_order=3,
) -> np.ndarray:
    """
    Compute the series at the given time points using the worm types to classify the points. The series is first corrected for incorrect segmentation, then median filtered to remove outliers, then smoothed using a Savitzky-Golay filter, and finally interpolated using b-splines.

    Parameters:
        series (np.ndarray): The time series.
        time (np.ndarray): The time points at which the series is to be computed.
        series_time (np.ndarray): The time points of the original series. If None, the time points are assumed to be the indices of the series.
        worm_types (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.
        lmbda (float, optional): The smoothing parameter for the Whittaker-Eilers smoothing. Default provides good results for our volume curves when series_time is in hours. (default: 0.0075)
        medfilt_window (int, optional): The window size for the median filter. (default: 5)
        bspline_order (int, optional): The order of the b-spline interpolation. (default: 3)

    Returns:
        np.ndarray: The series at the given time(s).
    """

    if np.all(np.isnan(series)):
        # handle time being a single value
        if np.isscalar(time):
            return np.full(1, np.nan)
        else:
            return np.full(time.shape, np.nan)

    # Smooth the series
    series = smooth_series_classified(
        series,
        series_time,
        worm_types,
        medfilt_window=medfilt_window,
        lmbda=lmbda,
    )

    # Interpolate the series using b-splines
    if series_time is None:
        series_time = np.arange(len(series))

    try:
        interpolated_series = interpolate.make_interp_spline(
            series_time, series, k=bspline_order
        )
    except Exception as e:
        print(f"Caught an exception while interpolating series: {e}")
        return np.full(time.shape, np.nan)

    return interpolated_series(time)


def rescale_series(series, time, ecdysis, worm_type, points=None, n_points=100):
    """
    Interpolates one or multiple series and the time to have n_points points in total per larval stages.

    Parameters:
        series (np.ndarray) : Array of shape (nb_of_worms / points, length of series) containing the series to interpolate.
        time (np.ndarray) : Array of shape (nb_of_worms / points, length of series) containing the time information.
        ecdysis (np.ndarray) : Array of shape (nb_of_worms / points, 5) containing the ecdysis times.
        worm_type (np.ndarray) : Array of shape (nb_of_worms / points, length of series) containing the worm type classification
        points (list) : List of points to consider for the interpolation. If None, all points are considered.
        n_points (int) : Number of points to interpolate the series to.

    Returns:
        all_points_interpolated_time (np.ndarray) : Array of shape (nb_of_worms / points, 4, n_points) containing the interpolated time information.
        all_points_interpolated_series (np.ndarray) : Array of shape (nb_of_worms / points, 4, n_points) containing the interpolated series.
    """
    if points is not None:
        series = series[points]
        time = time[points]
        worm_type = worm_type[points]
        ecdysis = ecdysis[points]

    # Interpolate the time and the series

    all_points_interpolated_time = []
    all_points_interpolated_series = []

    for point in range(series.shape[0]):
        series_point, time_point, ecdysis_point, worm_type_point = (
            series[point],
            time[point],
            ecdysis[point],
            worm_type[point],
        )
        (
            interpolated_time,
            interpolated_series,
        ) = interpolate_entire_development_classified(
            series_point,
            time_point,
            ecdysis_point,
            worm_type_point,
            n_points=n_points,
        )

        all_points_interpolated_time.append(interpolated_time)
        all_points_interpolated_series.append(interpolated_series)

    all_points_interpolated_time = np.array(all_points_interpolated_time)
    all_points_interpolated_series = np.array(all_points_interpolated_series)

    return all_points_interpolated_time, all_points_interpolated_series


def aggregate_interpolated_series(
    all_points_interpolated_series,
    all_points_interpolated_time,
    larval_stage_durations,
    aggregation="mean",
):
    """
    Aggregates the interpolated series and time information.

    Parameters:
        all_points_interpolated_series (np.ndarray) : Array of shape (nb_of_worms / points, 4, n_points) containing the interpolated series.
        all_points_interpolated_time (np.ndarray) : Array of shape (nb_of_worms / points, 4, n_points) containing the interpolated time information.
        larval_stage_durations (np.ndarray) : Array of shape (nb_of_worms / points, 4) containing the larval stage durations.
        aggregation (str) : Aggregation method to use. Can be 'mean' or 'median'.

    Returns:
        rescaled_time (np.ndarray) : Array of shape (4*n_points) containing the rescaled time information.
        aggregated_series (np.ndarray) : Array of shape (4*n_points) containing the aggregated series.
        std_series (np.ndarray) : Array of shape (4*n_points) containing the standard deviation of the series.
        ste_series (np.ndarray) : Array of shape (4*n_points) containing the standard error of the series.
    """
    if aggregation == "mean":
        aggregation_function = np.nanmean
    elif aggregation == "median":
        aggregation_function = np.nanmedian

    n_points = all_points_interpolated_time.shape[-1]

    aggregated_series = np.full((4, n_points), np.nan)
    std_series = np.full((4, n_points), np.nan)
    ste_series = np.full((4, n_points), np.nan)
    rescaled_time = np.full((4, n_points), np.nan)

    aggregated_larval_stage_durations = aggregation_function(
        larval_stage_durations, axis=0
    )

    for i in range(4):
        aggregated_series[i, :] = aggregation_function(
            all_points_interpolated_series[:, i, :], axis=0
        )
        std_series[i, :] = np.nanstd(all_points_interpolated_series[:, i, :], axis=0)
        ste_series[i, :] = np.nanstd(
            all_points_interpolated_series[:, i, :], axis=0
        ) / np.sqrt(np.sum(np.isfinite(all_points_interpolated_series[:, i, :])))

        beginning = (
            np.nansum(aggregated_larval_stage_durations[: i + 1])
            - aggregated_larval_stage_durations[i]
        )
        end = np.nansum(aggregated_larval_stage_durations[: i + 1])
        rescaled_time[i, :] = np.linspace(beginning, end, n_points)

    # flatten the arrays
    aggregated_series = aggregated_series.flatten()
    std_series = std_series.flatten()
    ste_series = ste_series.flatten()
    rescaled_time = rescaled_time.flatten()

    return rescaled_time, aggregated_series, std_series, ste_series


def rescale_and_aggregate(
    series,
    time,
    ecdysis,
    larval_stage_durations,
    worm_type,
    points=None,
    aggregation="mean",
    n_points=100,
):
    """
    Rescales and aggregates the series and time information.

    Parameters:
        series (np.ndarray) : Array of shape (nb_of_worms, length of series) containing the series to interpolate.
        time (np.ndarray) : Array of shape (nb_of_worms, length of series) containing the time information.
        ecdysis (np.ndarray) : Array of shape (nb_of_worms, 5) containing the ecdysis times.
        worm_type (np.ndarray) : Array of shape (nb_of_worms, length of series) containing the worm type classification
        points (list) : List of points to consider for the interpolation. If None, all points are considered.
        aggregation (str) : Aggregation method to use. Can be 'mean' or 'median'.
        n_points (int) : Number of points to interpolate the series to.

    Returns:
        rescaled_time (np.ndarray) : Array of shape (4*n_points) containing the rescaled time information.
        aggregated_series (np.ndarray) : Array of shape (4*n_points) containing the aggregated series.
        std_series (np.ndarray) : Array of shape (4*n_points) containing the standard deviation of the series.
        ste_series (np.ndarray) : Array of shape (4*n_points) containing the standard error of the series.
    """

    (
        all_points_interpolated_time,
        all_points_interpolated_series,
    ) = rescale_series(
        series, time, ecdysis, worm_type, points=points, n_points=n_points
    )
    (
        rescaled_time,
        aggregated_series,
        std_series,
        ste_series,
    ) = aggregate_interpolated_series(
        all_points_interpolated_series,
        all_points_interpolated_time,
        larval_stage_durations,
        aggregation=aggregation,
    )

    return rescaled_time, aggregated_series, std_series, ste_series



from towbintools.deep_learning.architectures import KeypointDetection1DModel
from towbintools.deep_learning.architectures import PretrainedSegmentationModel
from towbintools.deep_learning.architectures import SegmentationModel
from towbintools.deep_learning.utils.util import (
    get_input_channels_from_checkpoint,
)


def create_pretrained_segmentation_model(
    input_channels=3,
    n_classes=1,
    architecture="UnetPlusPlus",
    encoder="efficientnet-b4",
    pretrained_weights="image-micronet",
    normalization={"type": "percentile", "lo": 1, "hi": 99},
    learning_rate=1e-5,
    checkpoint_path=None,
    criterion=None,
):
    """
    Create a segmentation model with a pretrained encoder.

    Parameters:
        n_classes (int): The number of classes in the segmentation task.
        learning_rate (float): The learning rate for the optimizer.
        architecture (str): The architecture of the segmentation model.
        encoder (str): The encoder of the segmentation model.
        pretrained_weights (str): Dataset the encoder was trained on.
        normalization (dict): Parameters for the normalization.
        checkpoint_path (str): Path to a checkpoint file.
        criterion (torch.nn.Module): The loss function.

    Returns:
        PretrainedSegmentationModel: The segmentation model with a pretrained encoder.
    """
    if checkpoint_path is not None:
        model = PretrainedSegmentationModel.load_from_checkpoint(
            checkpoint_path, map_location="cpu"
        )
        # change the learning rate and normalization
        model.learning_rate = learning_rate
        model.normalization = normalization
        return model

    model = PretrainedSegmentationModel(
        input_channels=input_channels,
        n_classes=n_classes,
        learning_rate=learning_rate,
        architecture=architecture,
        encoder=encoder,
        pretrained_weights=pretrained_weights,
        normalization=normalization,
        criterion=criterion,
    )
    return model


def create_segmentation_model(
    architecture,
    input_channels,
    n_classes,
    normalization={"type": "percentile", "lo": 1, "hi": 99},
    learning_rate=1e-5,
    deep_supervision=False,
    checkpoint_path=None,
    criterion=None,
):
    """
    Create a segmentation model.

    Parameters:
        architecture (str): The architecture of the segmentation model.
        input_channels (int): The number of input channels.
        n_classes (int): The number of classes in the segmentation task.
        learning_rate (float): The learning rate for the optimizer.
        normalization (dict): Parameters for the normalization.
        deep_supervision (bool): Whether to use deep supervision or not.
        checkpoint_path (str): Path to a checkpoint file.
        criterion (torch.nn.Module): The loss function.

    Returns:
        SegmentationModel: The segmentation model.
    """

    if checkpoint_path is not None:
        model = SegmentationModel.load_from_checkpoint(checkpoint_path)
        return model

    model = SegmentationModel(
        architecture=architecture,
        input_channels=input_channels,
        n_classes=n_classes,
        learning_rate=learning_rate,
        normalization=normalization,
        deep_supervision=deep_supervision,
        criterion=criterion,
    )

    return model


def create_keypoint_detection_model(
    architecture,
    input_channels,
    n_classes,
    learning_rate=1e-4,
    checkpoint_path=None,
    criterion=None,
    activation="relu",
):
    """
    Create a keypoint detection model.

    Parameters:
        architecture (str): The architecture of the keypoint detection model.
        input_channels (int): The number of input channels.
        n_classes (int): The number of classes in the keypoint detection task.
        learning_rate (float): The learning rate for the optimizer.
        checkpoint_path (str): Path to a checkpoint file.
        criterion (torch.nn.Module): The loss function.
        activation (str): The activation function to use.

    Returns:
        KeypointDetection1DModel: The keypoint detection model.
    """

    if checkpoint_path is not None:
        model = KeypointDetection1DModel.load_from_checkpoint(checkpoint_path)
        return model

    model = KeypointDetection1DModel(
        input_channels,
        n_classes,
        learning_rate,
        architecture=architecture,
        activation=activation,
        criterion=criterion,
    )
    return model


def load_pretrained_model_from_checkpoint(checkpoint_path):
    """Load a pretrained segmentation model from a checkpoint. If the model cannot be loaded
    tries to load it with the default number of input channels.

    Parameters:
        checkpoint_path (str): Path to a checkpoint file.

    Returns:
        PretrainedSegmentationModel: The pretrained segmentation model.

    Raises:
        ValueError: If the model cannot be loaded from the checkpoint.
    """
    try:
        return PretrainedSegmentationModel.load_from_checkpoint(checkpoint_path)
    except Exception as e:
        try:
            return PretrainedSegmentationModel.load_from_checkpoint(
                checkpoint_path,
                input_channels=get_input_channels_from_checkpoint(checkpoint_path),
                pretrained_weights=None,
            )
        except Exception as e2:
            raise ValueError(
                f"Could not load model from checkpoint {checkpoint_path}. Error: {e} and {e2}"
            )


def load_scratch_segmentation_model_from_checkpoint(
    checkpoint_path, default_deep_supervision=False
):
    """Load a segmentation model from a checkpoint. If the model cannot be loaded
    tries to load it with the default number of input channels and deep supervision turned off.

    This function first tries to load the model as a PretrainedSegmentationModel. If that fails,
    it tries to load it as a SegmentationModel. If both attempts fail, it raises an error.

    Parameters:
        checkpoint_path (str): Path to a checkpoint file.

    Returns:
        SegmentationModel or PretrainedSegmentationModel: The segmentation model.

    Raises:
        ValueError: If the model cannot be loaded from the checkpoint.
    """
    try:
        return SegmentationModel.load_from_checkpoint(checkpoint_path)
    except Exception as e:
        try:
            return SegmentationModel.load_from_checkpoint(
                checkpoint_path,
                input_channels=get_input_channels_from_checkpoint(checkpoint_path),
                deep_supervision=default_deep_supervision,
            )
        except Exception as e2:
            raise ValueError(
                f"Could not load model from checkpoint {checkpoint_path}. Error: {e} and {e2}"
            )


def load_segmentation_model_from_checkpoint(checkpoint_path):
    """Load a segmentation model from a checkpoint.

    This function first tries to load the model as a PretrainedSegmentationModel. If that fails,
    it tries to load it as a SegmentationModel. If both attempts fail, it raises an error.

    Parameters:
        checkpoint_path (str): Path to a checkpoint file.

    Returns:
        SegmentationModel or PretrainedSegmentationModel: The segmentation model.

    Raises:
        ValueError: If the model cannot be loaded from the checkpoint.
    """

    try:
        return load_pretrained_model_from_checkpoint(checkpoint_path)
    except Exception as e:
        try:
            return load_scratch_segmentation_model_from_checkpoint(checkpoint_path)
        except Exception as e2:
            raise ValueError(
                f"Could not load model from checkpoint {checkpoint_path}. Error: {e} and {e2}"
            )


def load_keypoint_detection_model_from_checkpoint(checkpoint_path):
    """Load a keypoint detection model from a checkpoint.

    Parameters:
        checkpoint_path (str): Path to a checkpoint file.

    Returns:
        KeypointDetection1DModel: The keypoint detection model.

    Raises:
        ValueError: If the model cannot be loaded from the checkpoint.
    """
    try:
        return KeypointDetection1DModel.load_from_checkpoint(checkpoint_path)
    except Exception as e:
        raise ValueError(
            f"Could not load keypoint detection model from checkpoint {checkpoint_path}. Error: {e}"
        )



import albumentations as albu
import numpy as np
from albumentations.core.transforms_interface import DualTransform
from albumentations.core.transforms_interface import ImageOnlyTransform
from csbdeep.utils import normalize

from towbintools.foundation import image_handling


class NormalizeDataRange(ImageOnlyTransform):
    def __init__(self, always_apply=True, p=1.0):
        super().__init__(always_apply, p)

    def apply(self, img, **params):
        return (img - np.min(img)) / (np.max(img) - np.min(img))

    def get_transform_init_args_names(self):
        return ()


class NormalizeMeanStd(ImageOnlyTransform):
    def __init__(self, mean, std, always_apply=True, p=1.0):
        super().__init__(always_apply, p)
        self.mean = mean
        self.std = std

    def apply(self, img, **params):
        return (img - self.mean) / self.std

    def get_transform_init_args_names(self):
        return ("mean", "std")


class NormalizePercentile(ImageOnlyTransform):
    def __init__(self, lo, hi, axis=None, always_apply=True, p=1.0):
        super().__init__(always_apply, p)
        self.lo = lo
        self.hi = hi
        self.axis = axis

    def apply(self, img, **params):
        return normalize(img, self.lo, self.hi, axis=self.axis)

    def get_transform_init_args_names(self):
        return ("lo", "hi")


class EnforceNChannels(ImageOnlyTransform):
    def __init__(self, n_channels, always_apply=True, p=1.0):
        super().__init__(always_apply, p)
        self.n_channels = n_channels

    def apply(self, img, **params):
        return enforce_n_channels(img, self.n_channels)

    def get_transform_init_args_names(self):
        return ("n_channels",)


class CustomFlip(DualTransform):
    """Flip the input image horizontally or vertically with a given probability. Works well with images ordered in the OME-TIFF way."""

    def __init__(self, always_apply=False, p=0.5):
        super().__init__(always_apply=always_apply, p=p)
        self.axis = np.random.choice([-1, -2])

    def apply(self, img, **params):
        return np.flip(img, axis=self.axis)

    def apply_to_mask(self, img, **params):
        return np.flip(img, axis=self.axis)

    def get_transform_init_args_names(self):
        return ()


class CustomRotate90(DualTransform):
    """Rotate the input image by 90 degrees."""

    def __init__(self, always_apply=False, p=0.5):
        super().__init__(always_apply=always_apply, p=p)
        self.axis = np.random.choice([-1, -2])
        self.k = np.random.choice([1, 2, 3])

    def apply(self, img, **params):
        return np.rot90(img, k=self.k, axes=(-2, -1))

    def apply_to_mask(self, img, **params):
        return np.rot90(img, k=self.k, axes=(-2, -1))

    def get_transform_init_args_names(self):
        return ()


def get_training_augmentation(normalization_type, **kwargs):
    train_transform = [
        CustomFlip(p=0.75),
        CustomRotate90(p=0.75),
        albu.Defocus(p=0.5),
        albu.RandomGamma(p=0.5),
    ]

    if normalization_type == "data_range":
        train_transform.append(NormalizeDataRange())
    elif normalization_type == "mean_std":
        train_transform.append(NormalizeMeanStd(kwargs["mean"], kwargs["std"]))
    elif normalization_type == "percentile":
        try:
            train_transform.append(
                NormalizePercentile(kwargs["lo"], kwargs["hi"], kwargs["axis"])
            )
        except KeyError:
            train_transform.append(NormalizePercentile(kwargs["lo"], kwargs["hi"]))

    enforce_n_channels = kwargs.get("enforce_n_channels", None)

    if enforce_n_channels is not None:
        train_transform.append(EnforceNChannels(enforce_n_channels))

    return albu.Compose(train_transform)


def get_prediction_augmentation(normalization_type, **kwargs):
    prediction_transform = []

    if normalization_type == "data_range":
        prediction_transform.append(NormalizeDataRange())
    elif normalization_type == "mean_std":
        prediction_transform.append(NormalizeMeanStd(kwargs["mean"], kwargs["std"]))
    elif normalization_type == "percentile":
        try:
            prediction_transform.append(
                NormalizePercentile(kwargs["lo"], kwargs["hi"], kwargs["axis"])
            )
        except KeyError:
            prediction_transform.append(NormalizePercentile(kwargs["lo"], kwargs["hi"]))

    enforce_n_channels = kwargs.get("enforce_n_channels", None)

    if enforce_n_channels is not None:
        prediction_transform.append(EnforceNChannels(enforce_n_channels))

    return albu.Compose(prediction_transform)


def get_prediction_augmentation_from_model(model, enforce_n_channels=None):
    normalization_type = model.normalization["type"]
    normalization_params = model.normalization

    if normalization_type == "percentile":
        try:
            preprocessing_fn = get_prediction_augmentation(
                normalization_type=normalization_type,
                lo=normalization_params["lo"],
                hi=normalization_params["hi"],
                axis=normalization_params["axis"],
                enforce_n_channels=enforce_n_channels,
            )
        except KeyError:
            preprocessing_fn = get_prediction_augmentation(
                normalization_type=normalization_type,
                lo=normalization_params["lo"],
                hi=normalization_params["hi"],
                enforce_n_channels=enforce_n_channels,
            )
    elif normalization_type == "mean_std":
        preprocessing_fn = get_prediction_augmentation(
            normalization_type=normalization_type,
            mean=normalization_params["mean"],
            std=normalization_params["std"],
            enforce_n_channels=enforce_n_channels,
        )
    elif normalization_type == "data_range":
        preprocessing_fn = get_prediction_augmentation(
            normalization_type=normalization_type,
            enforce_n_channels=enforce_n_channels,
        )
    else:
        preprocessing_fn = get_prediction_augmentation(
            normalization_type=normalization_type,
            enforce_n_channels=enforce_n_channels,
        )

    return preprocessing_fn


def get_mean_and_std(image_path):
    image = image_handling.read_tiff_file(image_path, [2])
    return np.mean(image), np.std(image)


# def enforce_n_channels(image, n_channels):
#     if not isinstance(image, torch.Tensor):
#         image = torch.tensor(image, dtype=torch.float32)

#     assert (
#         len(image.shape) <= 3
#     ), "Currently, multichannel zstacks are not supported"

#     if len(image.shape) == 2:
#         image = image.unsqueeze(0)

#     current_channels = image.shape[0]

#     # Assuming grayscale_img has a shape of (C, H, W)
#     if current_channels == n_channels:
#         return image

#     if current_channels > n_channels:
#         raise ValueError(
#             f"The image has more channels than the specified number of channels ({n_channels})"
#         )

#     if n_channels % current_channels == 0:
#         return image.repeat((n_channels // current_channels, 1, 1))

#     else:
#         # First repeat the maximum number of times that divides evenly
#         base_repeats = n_channels // current_channels
#         remaining_channels = n_channels % current_channels

#         # Create the base repeated tensor
#         repeated = image.repeat((base_repeats, 1, 1))

#         # Add the remaining channels by selecting from the beginning
#         remaining = image[:remaining_channels]

#         # Concatenate along the channel dimension
#         return torch.cat([repeated, remaining], dim=0)


def enforce_n_channels(image, n_channels):
    if not isinstance(image, np.ndarray):
        image = np.array(image, dtype=np.float32)

    assert len(image.shape) <= 3, "Currently, multichannel zstacks are not supported"

    # Add channel dimension if necessary
    if len(image.shape) == 2:
        image = np.expand_dims(image, axis=0)

    current_channels = image.shape[0]

    # Return if already correct number of channels
    if current_channels == n_channels:
        return image

    if current_channels > n_channels:
        raise ValueError(
            f"The image has more channels than the specified number of channels ({n_channels})"
        )

    if n_channels % current_channels == 0:
        # Use np.tile instead of torch.repeat
        return np.tile(image, (n_channels // current_channels, 1, 1))
    else:
        # First repeat the maximum number of times that divides evenly
        base_repeats = n_channels // current_channels
        remaining_channels = n_channels % current_channels

        # Create the base repeated array
        repeated = np.tile(image, (base_repeats, 1, 1))

        # Add the remaining channels by selecting from the beginning
        remaining = image[:remaining_channels]

        # Concatenate along the channel dimension
        return np.concatenate([repeated, remaining], axis=0)



from .archs import Unet
from .archs import UnetPlusPlus
from .models import KeypointDetection1DModel
from .models import PretrainedSegmentationModel
from .models import SegmentationModel

__all__ = [
    "Unet",
    "UnetPlusPlus",
    "PretrainedSegmentationModel",
    "SegmentationModel",
    "Unet1D",
    "AttentionUnet1D",
    "UnetPlusPlus1D",
    "KeypointDetection1DModel",
]



import pytorch_lightning as pl
import segmentation_models_pytorch as smp
import timm
import torch
import torch.nn as nn
from torchmetrics.classification import BinaryF1Score
from torchmetrics.classification import MulticlassF1Score

from .archs import AttentionUnet1D
from .archs import Unet
from .archs import Unet1D
from .archs import UnetPlusPlus
from .archs import UnetPlusPlus1D
from towbintools.deep_learning.utils.loss import FocalTverskyLoss
from towbintools.deep_learning.utils.loss import PeakWeightedMSELoss


class PretrainedClassificationModel(pl.LightningModule):
    """Pytorch Lightning Module for training a classification model with a pretrained weights. The model and weights are loaded from the pretrained_microscopy_models package.
    Because of the way they were pretrained, input images are required to have 3 channels.

    Parameters:
            architecture (str): The architecture of the classification model.
            input_channels (int): The number of input channels.
            n_classes (int): The number of classes in the segmentation task.
            learning_rate (float): The learning rate for the optimizer.
            normalization (dict): Parameters for the normalization.
    """

    def __init__(
        self,
        architecture,
        input_channels,
        n_classes,
        learning_rate,
        normalization,
    ):
        super().__init__()
        model = timm.create_model(
            architecture,
            pretrained=True,
            num_classes=n_classes,
            in_chans=input_channels,
        )

        self.model = model
        self.learning_rate = learning_rate
        self.n_classes = n_classes
        if n_classes == 2:
            self.criterion = nn.BCELoss()
            self.f1_score = BinaryF1Score()
        else:
            self.criterion = nn.CrossEntropyLoss()
            self.f1_score = MulticlassF1Score(num_classes=n_classes)

        self.normalization = normalization
        self.save_hyperparameters()

    def forward(self, x):
        if self.n_classes == 2:
            return torch.sigmoid(self.model(x))
        else:
            return torch.softmax(self.model(x), dim=1)

    def log_tb_images(self, viz_batch) -> None:
        # Get tensorboard logger
        tb_logger = None
        for logger in self.trainer.loggers:
            if isinstance(logger, pl.loggers.TensorBoardLogger):
                tb_logger = logger.experiment
                break

        if tb_logger is None:
            raise ValueError("TensorBoard Logger not found")

    def training_step(self, batch):
        x, y = batch
        y_hat = self.forward(x).squeeze()
        if y_hat.dim() == 0:
            y_hat = y_hat.unsqueeze(0)
        y = y.to(torch.float)
        loss = self.criterion(y_hat, y)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "train_f1_score",
            f1_score,
            on_step=False,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

        return loss

    def validation_step(self, batch):
        x, y = batch
        y_hat = self.forward(x).squeeze()
        if y_hat.dim() == 0:
            y_hat = y_hat.unsqueeze(0)
        y = y.to(torch.float)
        loss = self.criterion(y_hat, y)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "val_f1_score",
            f1_score,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer


class PretrainedSegmentationModel(pl.LightningModule):
    """Pytorch Lightning Module for training a segmentation model with a pretrained encoder. The encoder is loaded from the pretrained_microscopy_models package.
    This model automatically contains an activation layer at the end addapted to the number of classes.

    Parameters:
            input_channels (int): The number of input channels.
            n_classes (int): The number of classes in the segmentation task.
            learning_rate (float): The learning rate for the optimizer.
            architecture (str): The architecture of the segmentation model.
            encoder (str): The encoder of the segmentation model.
            pretrained_weights (str): Dataset the encoder was trained on.
            normalization (dict): Parameters for the normalization.
            criterion (torch.nn.Module): The loss function to use for training. (default: FocalTverskyLoss)
            ignore_index (int): Index to ignore in the loss calculation and F1Score.

    """

    def __init__(
        self,
        input_channels,
        n_classes,
        learning_rate,
        architecture,
        encoder,
        pretrained_weights,
        normalization,
        criterion=None,
        ignore_index=None,
    ):
        super().__init__()

        if n_classes == 1:
            activation = "sigmoid"
        else:
            activation = "softmax"

        model = smp.create_model(
            arch=architecture,
            encoder_name=encoder,
            encoder_weights=pretrained_weights,
            in_channels=input_channels,
            classes=n_classes,
            activation=activation,
        )

        self.model = model
        self.learning_rate = learning_rate
        self.ignore_index = ignore_index

        if criterion is None:
            if n_classes == 1:
                self.criterion = FocalTverskyLoss(ignore_index=self.ignore_index)
            else:
                self.criterion = nn.CrossEntropyLoss(ignore_index=self.ignore_index)
        else:
            self.criterion = criterion

        try:
            if n_classes == 1:
                self.f1_score = BinaryF1Score(ignore_index=self.criterion.ignore_index)
            else:
                self.f1_score = MulticlassF1Score(
                    num_classes=n_classes,
                    ignore_index=self.criterion.ignore_index,
                )
        except Exception as e:
            print(f"Criterion does not support ignore_index: {e}")
            if n_classes == 1:
                self.f1_score = BinaryF1Score()
            else:
                self.f1_score = MulticlassF1Score(num_classes=n_classes)

        self.normalization = normalization
        self.save_hyperparameters(ignore=["criterion"])

    def forward(self, x):
        return self.model(x)

    def training_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        loss = self.criterion(y_hat, y)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "train_f1_score",
            f1_score,
            on_step=False,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

        return loss

    def validation_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        loss = self.criterion(y_hat, y)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "val_f1_score",
            f1_score,
            on_step=True,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

    def predict_step(self, batch):
        x = batch

        pred = self.model(x)  # prediction already has softmax / sigmoid applied

        # binarize predictions
        if self.n_classes == 1:
            pred = pred > 0.5
        else:
            pred = torch.argmax(pred, dim=1)

        return pred

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer


class SegmentationModel(pl.LightningModule):
    def __init__(
        self,
        architecture,
        input_channels,
        n_classes,
        learning_rate,
        normalization,
        deep_supervision=False,
        criterion=None,
        ignore_index=None,
    ):
        """Pytorch Lightning Module for training a segmentation model.

        Parameters:
            architecture (str): The architecture of the segmentation model.
            input_channels (int): The number of input channels.
            n_classes (int): The number of classes in the segmentation task.
            learning_rate (float): The learning rate for the optimizer.
            normalization (dict): Parameters for the normalization.
            deep_supervision (bool): Whether to use deep supervision or not.
            criterion (torch.nn.Module): The loss function to use for training. Default is FocalTverskyLoss.
            ignore_index (int): Index to ignore in the loss calculation and F1Score.

        """

        super().__init__()
        if architecture == "Unet":
            model = Unet(num_classes=n_classes, input_channels=input_channels)
        elif architecture == "UnetPlusPlus":
            model = UnetPlusPlus(
                num_classes=n_classes,
                input_channels=input_channels,
                deep_supervision=deep_supervision,
            )
        else:
            raise ValueError(
                f"Architecture {architecture} not implemented. Implemented architectures are: Unet, UnetPlusPlus"
            )
        self.model = model
        self.learning_rate = learning_rate
        self.ignore_index = ignore_index

        if criterion is None:
            if n_classes == 1:
                self.criterion = FocalTverskyLoss(ignore_index=self.ignore_index)
            else:
                self.criterion = nn.CrossEntropyLoss(ignore_index=self.ignore_index)
        else:
            self.criterion = criterion

        try:
            if n_classes == 1:
                self.f1_score = BinaryF1Score(ignore_index=self.criterion.ignore_index)
            else:
                self.f1_score = MulticlassF1Score(
                    num_classes=n_classes,
                    ignore_index=self.criterion.ignore_index,
                )
        except Exception as e:
            print(f"Criterion does not support ignore_index: {e}")
            if n_classes == 1:
                self.f1_score = BinaryF1Score()
            else:
                self.f1_score = MulticlassF1Score(num_classes=n_classes)

        if n_classes == 1:
            self.activation = nn.Sigmoid()
        else:
            self.activation = nn.Softmax(dim=1)

        self.normalization = normalization
        self.save_hyperparameters()

    def forward(self, x):
        y = self.model(x)
        return self.activation(y)

    def training_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        y_hat = self.activation(y_hat)
        loss = self.criterion(y_hat, y)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "train_f1_score",
            f1_score,
            on_step=False,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )
        return loss

    def validation_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        y_hat = self.activation(y_hat)
        loss = self.criterion(y_hat, y)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

        f1_score = self.f1_score(y_hat, y)
        self.log(
            "val_f1_score",
            f1_score,
            on_step=True,
            on_epoch=True,
            logger=True,
            sync_dist=True,
        )

    def predict_step(self, batch):
        x = batch
        pred = self.model(x)
        pred = self.activation(pred)

        # binarize predictions
        if self.n_classes == 1:
            pred = pred > 0.5
        else:
            pred = torch.argmax(pred, dim=1)

        return pred

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer


class KeypointDetection1DModel(pl.LightningModule):
    def __init__(
        self,
        input_channels,
        n_classes,
        learning_rate,
        architecture="UnetPlusPlus",
        activation="sigmoid",
        criterion=None,
    ):
        """Pytorch Lightning Module for 1D Keypoint Detection using a U-Net architecture.

        Parameters:
            input_channels (int): The number of input channels.
            n_classes (int): The number of classes in the segmentation task.
            learning_rate (float): The learning rate for the optimizer.
            architecture (str): The architecture of the segmentation model. (default: "Unet")
            criterion (torch.nn.Module): The loss function to use for training. (default: nn.MSELoss)
            activation (str): The activation function to use at the end of the model. (default: "sigmoid")
        """

        super().__init__()

        if architecture == "Unet":
            model = Unet1D(num_classes=n_classes, input_channels=input_channels)
        elif architecture == "AttentionUnet":
            model = AttentionUnet1D(
                num_classes=n_classes, input_channels=input_channels
            )
        elif architecture == "UnetPlusPlus":
            model = UnetPlusPlus1D(num_classes=n_classes, input_channels=input_channels)
        else:
            raise ValueError(f"Unsupported architecture: {architecture}")

        self.model = model
        self.learning_rate = learning_rate

        # Set a default ignore_index if not present
        if not hasattr(self, "ignore_index"):
            self.ignore_index = -100

        if criterion is None:
            # self.criterion = nn.MSELoss()
            # self.criterion = nn.L1Loss()
            self.criterion = PeakWeightedMSELoss()

        else:
            self.criterion = criterion

        if activation == "relu":
            self.activation = nn.ReLU(inplace=True)
        elif activation == "leaky_relu":
            self.activation = nn.LeakyReLU(inplace=True)
        elif activation == "sigmoid":
            self.activation = nn.Sigmoid()
        elif activation == "none":
            self.activation = nn.Identity()
        else:
            raise ValueError(f"Unsupported activation function: {activation}")

        self.save_hyperparameters()

    def forward(self, x):
        y = self.model(x)
        return self.activation(y)

    def training_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        y_hat = self.activation(y_hat)
        loss = self.criterion(y_hat, y)
        self.log(
            "train_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )
        return loss

    def validation_step(self, batch):
        x, y = batch
        y_hat = self.model(x)
        y_hat = self.activation(y_hat)
        loss = self.criterion(y_hat, y)
        self.log(
            "val_loss",
            loss,
            on_step=True,
            on_epoch=True,
            prog_bar=True,
            logger=True,
            sync_dist=True,
        )

    def predict_step(self, batch):
        x = batch
        pred = self.model(x)
        return pred

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.learning_rate)
        return optimizer



from .quantification_tools import compute_background_fluorescence
from .quantification_tools import compute_fluorescence_in_mask

__all__ = [
    "compute_background_fluorescence",
    "compute_fluorescence_in_mask",
]



from .segmentation_tools import deep_learning_segmentation
from .segmentation_tools import edge_based_segmentation
from .segmentation_tools import segment_image

__all__ = [
    "deep_learning_segmentation",
    "edge_based_segmentation",
    "segment_image",
]



import datetime
import os

import numpy as np
import pandas as pd
import torch
from joblib import delayed
from joblib import Parallel
from pytorch_toolbelt import inference
from sklearn.model_selection import train_test_split
from torch.utils.data import DataLoader
from torch.utils.data import Dataset

from towbintools.data_analysis.time_series import crop_series_to_length
from towbintools.data_analysis.time_series import pad_series_to_length
from towbintools.deep_learning.utils.augmentation import (
    get_prediction_augmentation,
)
from towbintools.deep_learning.utils.augmentation import (
    get_training_augmentation,
)
from towbintools.deep_learning.utils.util import get_closest_lower_multiple
from towbintools.deep_learning.utils.util import get_closest_upper_multiple
from towbintools.foundation import image_handling


class TiledSegmentationDataset(Dataset):
    def __init__(
        self,
        dataset,
        image_slicers,
        channels,
        mask_column="mask",
        image_column="image",
        transform=None,
    ):
        self.images = dataset[image_column].values.tolist()
        self.ground_truth = dataset[mask_column].values.tolist()
        self.channels = channels
        self.image_slicers = image_slicers
        self.transform = transform

    def __len__(self):
        return len(self.images)

    def __getitem__(self, i):
        img = image_handling.read_tiff_file(self.images[i], [self.channels])
        mask = image_handling.read_tiff_file(self.ground_truth[i])

        if self.transform is not None:
            transformed = self.transform(image=img, mask=mask)
            img = transformed["image"]
            mask = transformed["mask"]

        slicer = self.image_slicers[img.shape]
        tiles = slicer.split(img)

        if len(tiles[0].shape) == 2:
            tiles = [tile[np.newaxis, ...] for tile in tiles]

        tiles_ground_truth = slicer.split(mask)

        selected_tile = np.random.randint(0, len(tiles))
        img = tiles[selected_tile]
        mask = tiles_ground_truth[selected_tile]
        mask = mask[np.newaxis, ...]

        return img.astype(np.float32), mask


class SegmentationDataset(Dataset):
    def __init__(
        self,
        dataset,
        channels,
        mask_column="mask",
        image_column="image",
        transform=None,
        enforce_divisibility_by=32,
        pad_or_crop="pad",
        mask_pad_value=-1,
    ):
        self.images = dataset[image_column].values.tolist()
        self.ground_truth = dataset[mask_column].values.tolist()
        self.channels = channels
        self.transform = transform
        self.enforce_divisibility_by = enforce_divisibility_by
        if pad_or_crop not in ["pad", "crop"]:
            raise ValueError("pad_or_crop must be either 'pad' or 'crop'")

        if pad_or_crop == "pad":
            self.resize_function = image_handling.pad_to_dim_equally
            self.mask_resize_function = (
                lambda dim, new_dim_x, new_dim_y: image_handling.pad_to_dim_equally(
                    dim, new_dim_x, new_dim_y, pad_value=mask_pad_value
                )
            )
            self.multiplier_function = get_closest_upper_multiple
        else:
            self.resize_function = image_handling.crop_to_dim_equally
            self.mask_resize_function = image_handling.crop_to_dim_equally
            self.multiplier_function = get_closest_lower_multiple

    def __len__(self):
        return len(self.images)

    def __getitem__(self, i):
        img = image_handling.read_tiff_file(self.images[i], [self.channels])
        mask = image_handling.read_tiff_file(self.ground_truth[i])

        if self.transform is not None:
            transformed = self.transform(image=img, mask=mask)
            img = transformed["image"]
            mask = transformed["mask"]

        if self.enforce_divisibility_by is not None:
            dim_x, dim_y = img.shape[-2:]

            if (
                dim_x % self.enforce_divisibility_by != 0
                or dim_y % self.enforce_divisibility_by != 0
            ):
                new_dim_x = self.multiplier_function(
                    dim_x, self.enforce_divisibility_by
                )
                new_dim_y = self.multiplier_function(
                    dim_y, self.enforce_divisibility_by
                )

                img = self.resize_function(img, new_dim_x, new_dim_y)
                mask = self.mask_resize_function(mask, new_dim_x, new_dim_y)

        if len(img.shape) == 2:
            img = img[np.newaxis, ...]
        mask = mask[np.newaxis, ...]

        return img.astype(np.float32), mask


class SegmentationPredictionDataset(Dataset):
    def __init__(
        self,
        image_paths,
        channels,
        transform=None,
        enforce_divisibility_by=32,
        pad_or_crop="pad",
    ):
        self.images = image_paths
        self.channels = channels
        self.transform = transform
        self.enforce_divisibility_by = enforce_divisibility_by
        if pad_or_crop not in ["pad", "crop"]:
            raise ValueError("pad_or_crop must be either 'pad' or 'crop'")

        self.pad_or_crop = pad_or_crop

        if pad_or_crop == "pad":
            self.resize_function = image_handling.pad_to_dim_equally
            self.multiplier_function = get_closest_upper_multiple
        else:
            self.resize_function = image_handling.crop_to_dim_equally
            self.multiplier_function = get_closest_lower_multiple

    def __len__(self):
        return len(self.images)

    def __getitem__(self, i):
        img_path = self.images[i]
        img = image_handling.read_tiff_file(img_path, [self.channels])

        if self.transform is not None:
            transformed = self.transform(image=img)
            img = transformed["image"]

        if len(img.shape) == 2:
            img = img[np.newaxis, ...]

        return img_path, img.astype(np.float32), img.shape

    def collate_fn(self, batch):
        img_paths, imgs, original_shapes = zip(*batch)

        if self.enforce_divisibility_by is None:
            return img_paths, imgs, original_shapes

        # find the maximum dimensions in the batch
        if self.pad_or_crop == "pad":
            dim_x = max([shape[-2] for shape in original_shapes])
            dim_y = max([shape[-1] for shape in original_shapes])
        else:
            dim_x = min([shape[-2] for shape in original_shapes])
            dim_y = min([shape[-1] for shape in original_shapes])

        # find the new dimensions
        new_dim_x = self.multiplier_function(dim_x, self.enforce_divisibility_by)
        new_dim_y = self.multiplier_function(dim_y, self.enforce_divisibility_by)

        # resize the images
        resized_images = []
        for img in imgs:
            resized_images.append(self.resize_function(img, new_dim_x, new_dim_y))
        resized_images = np.array(resized_images, dtype=np.float32)

        return img_paths, resized_images, original_shapes


class ClassificationDataset(Dataset):
    def __init__(
        self,
        dataset,
        channels,
        n_classes,
        class_column="class",
        image_column="image",
        transform=None,
    ):
        self.images = dataset[image_column].values.tolist()
        self.ground_truth = dataset[class_column].values.astype(float).tolist()
        self.channels = channels
        self.transform = transform

        # convert ground truth to one-hot encoding
        if n_classes > 2:
            self.ground_truth = np.eye(n_classes)[self.ground_truth]

    def __len__(self):
        return len(self.images)

    def __getitem__(self, i):
        img = image_handling.read_tiff_file(self.images[i], [self.channels])
        class_value = self.ground_truth[i]

        if self.transform is not None:
            transformed = self.transform(image=img)
            img = transformed["image"]

        if len(img.shape) == 2:
            img = img[np.newaxis, ...]

        return img.astype(np.float32), class_value


class KeypointDetection1DTrainingDataset(Dataset):
    def __init__(
        self,
        inputs,
        targets,
        enforce_divisibility_by=32,
        resize_method="pad",
    ):
        self.input_series = inputs
        self.targets = targets
        self.enforce_divisibility_by = enforce_divisibility_by

        self.resize_method = resize_method
        if resize_method == "pad":
            self.resize_function = pad_series_to_length
            self.target_resize_function = pad_series_to_length
            self.multiplier_function = get_closest_upper_multiple
        elif resize_method == "crop":
            self.resize_function = crop_series_to_length
            self.target_resize_function = crop_series_to_length
            self.multiplier_function = get_closest_lower_multiple

    def __len__(self):
        return len(self.input_series)

    def __getitem__(self, i):
        series = self.input_series[i]
        target = self.targets[i]

        if series.ndim == 1:
            series = series.reshape(1, -1)
        if target.ndim == 1:
            target = target.reshape(1, -1)

        return series.astype(np.float32), target.astype(np.float32), series.shape

    def collate_fn(self, batch):
        series, targets, original_shapes = zip(*batch)

        if self.enforce_divisibility_by is None:
            return series, targets, original_shapes

        if self.resize_method == "crop":
            target_length = min([shape[-1] for shape in original_shapes])
        else:
            target_length = max([shape[-1] for shape in original_shapes])

        new_length = self.multiplier_function(
            target_length, self.enforce_divisibility_by
        )

        # resize the images
        resized_series = []
        for series_i in series:
            resized_series_i = self.resize_function(series_i, new_length)
            resized_series.append(resized_series_i)

        resized_series = np.array(resized_series, dtype=np.float32)

        resized_targets = []
        for target_i in targets:
            resized_target_i = self.target_resize_function(target_i, new_length)
            resized_targets.append(resized_target_i)
        resized_targets = np.array(resized_targets, dtype=np.float32)

        # remove series if they contain NaN values
        valid_series_index = [
            i
            for i, series_i in enumerate(resized_series)
            if not np.any(np.isnan(series_i))
        ]
        resized_series = resized_series[valid_series_index]
        resized_targets = resized_targets[valid_series_index]
        original_shapes = [original_shapes[i] for i in valid_series_index]

        if type(resized_series) is np.ndarray:
            resized_series = torch.tensor(resized_series, dtype=torch.float32)
        if type(resized_targets) is np.ndarray:
            resized_targets = torch.tensor(resized_targets, dtype=torch.float32)

        return resized_series, resized_targets


class KeypointDetection1DPredictionDataset(Dataset):
    def __init__(
        self,
        inputs,
        enforce_divisibility_by=32,
        resize_method="pad",
    ):
        self.input_series = inputs
        self.enforce_divisibility_by = enforce_divisibility_by

        self.resize_method = resize_method
        if resize_method == "pad":
            self.resize_function = pad_series_to_length
            self.multiplier_function = get_closest_upper_multiple
        elif resize_method == "crop":
            self.resize_function = crop_series_to_length
            self.multiplier_function = get_closest_lower_multiple

    def __len__(self):
        return len(self.input_series)

    def __getitem__(self, i):
        series = self.input_series[i]

        if series.ndim == 1:
            series = series.reshape(1, -1)

        return series.astype(np.float32), series.shape

    def collate_fn(self, batch):
        series, original_shapes = zip(*batch)

        if self.enforce_divisibility_by is None:
            return series, original_shapes

        if self.resize_method == "crop":
            target_length = min([shape[-1] for shape in original_shapes])
        else:
            target_length = max([shape[-1] for shape in original_shapes])

        new_length = self.multiplier_function(
            target_length, self.enforce_divisibility_by
        )

        # resize the images
        resized_series = []
        for series_i in series:
            resized_series_i = self.resize_function(series_i, new_length)
            resized_series.append(resized_series_i)

        resized_series = np.array(resized_series, dtype=np.float32)

        # any series containing NaN would cause the batch to fail, so we replace them with zeros
        invalid_series_index = [
            i for i, series_i in enumerate(resized_series) if np.any(np.isnan(series_i))
        ]
        for i in invalid_series_index:
            resized_series[i] = np.zeros(new_length, dtype=np.float32)

        if type(resized_series) is np.ndarray:
            resized_series = torch.tensor(resized_series, dtype=torch.float32)

        return resized_series, invalid_series_index, original_shapes


def split_dataset(dataframe, validation_size, test_size):
    # Load the dataset
    if isinstance(dataframe, str):
        dataframe = pd.read_csv(dataframe)

    # Ensure the sizes are valid
    if validation_size + test_size >= 1.0:
        raise ValueError(
            "The sum of validation_size and test_size should be less than 1."
        )

    # Calculate the size of the temporary set (validation + test)
    temp_size = validation_size + test_size

    # Split the dataframe into training and temporary set
    train_dataframe, temp_dataframe = train_test_split(
        dataframe, test_size=temp_size, random_state=42
    )

    # Calculate the proportion of validation and test set relative to the temporary set
    validation_proportion = validation_size / temp_size

    # Split the temporary set into validation and test sets
    validation_dataframe, test_dataframe = train_test_split(
        temp_dataframe, test_size=1 - validation_proportion, random_state=42
    )

    return train_dataframe, validation_dataframe, test_dataframe


def create_segmentation_training_dataframes(
    image_directories,
    mask_directories,
    save_dir,
    validation_set_ratio=0.25,
    test_set_ratio=0.1,
):
    if not isinstance(image_directories, list):
        image_directories = [image_directories]
    if not isinstance(mask_directories, list):
        mask_directories = [mask_directories]

    images = []
    masks = []
    for image_directory, mask_directory in zip(image_directories, mask_directories):
        images.extend(
            sorted(
                [
                    os.path.join(image_directory, file)
                    for file in os.listdir(image_directory)
                ]
            )
        )
        masks.extend(
            sorted(
                [
                    os.path.join(mask_directory, file)
                    for file in os.listdir(mask_directory)
                ]
            )
        )

    assert len(images) == len(
        masks
    ), "The number of images and masks in the directories must be equal"
    dataframe = pd.DataFrame({"image": images, "mask": masks})

    training_dataframe, validation_dataframe, test_dataframe = split_dataset(
        dataframe, validation_set_ratio, test_set_ratio
    )

    # backup the training and validation dataframes
    database_backup_dir = os.path.join(save_dir, "database_backup")
    current_date = datetime.datetime.now().strftime("%Y%m%d")
    os.makedirs(database_backup_dir, exist_ok=True)

    training_dataframe.to_csv(
        os.path.join(database_backup_dir, f"training_dataframe_{current_date}.csv"),
        index=False,
    )
    validation_dataframe.to_csv(
        os.path.join(database_backup_dir, f"validation_dataframe_{current_date}.csv"),
        index=False,
    )
    test_dataframe.to_csv(
        os.path.join(database_backup_dir, f"test_dataframe_{current_date}.csv"),
        index=False,
    )

    return training_dataframe, validation_dataframe


def get_unique_shapes_from_tiffs(image_paths=list[str]) -> np.ndarray:
    """
    Get unique shapes from a list of TIFF images in parallel.

    Parameters:
        image_paths (List[str]): List of image paths to extract shapes from

    Returns:
        np.ndarray: Unique image shapes found in the dataframe
    """

    shapes = Parallel(n_jobs=-1)(
        delayed(image_handling.get_shape_from_tiff)(image_path)
        for image_path in image_paths
    )

    valid_shapes = [shape for shape in shapes if shape is not None]

    if len(valid_shapes) == 0:
        raise ValueError("No valid shapes found in the dataframe")

    return np.unique(valid_shapes, axis=0)


def create_segmentation_dataloaders(
    training_dataframe,
    validation_dataframe,
    channels,
    batch_size=5,
    num_workers=32,
    pin_memory=True,
    train_on_tiles=True,
    tiler_params=None,
    training_transform=None,
    validation_transform=None,
):
    if not train_on_tiles:
        train_loader = DataLoader(
            SegmentationDataset(
                training_dataframe,
                channels=channels,
                mask_column="mask",
                image_column="image",
                transform=training_transform,
            ),
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )
        val_loader = DataLoader(
            SegmentationDataset(
                validation_dataframe,
                channels=channels,
                mask_column="mask",
                image_column="image",
                transform=validation_transform,
            ),
            batch_size=batch_size,
            shuffle=False,
            num_workers=num_workers,
            pin_memory=pin_memory,
        )
        return train_loader, val_loader

    assert (
        tiler_params is not None
    ), "If train_on_tiles is True, tiler_params must be provided"

    image_paths = training_dataframe["image"].values.tolist()

    unique_shapes = get_unique_shapes_from_tiffs(image_paths)
    image_slicers = {
        tuple(shape): inference.ImageSlicer(
            shape, tiler_params["tile_size"], tiler_params["tile_step"]
        )
        for shape in unique_shapes
    }

    if training_transform is None:
        training_transform = get_training_augmentation("percentile", lo=1, hi=99)
    if validation_transform is None:
        validation_transform = get_prediction_augmentation("percentile", lo=1, hi=99)

    train_loader = DataLoader(
        TiledSegmentationDataset(
            training_dataframe,
            image_slicers,
            channels=channels,
            mask_column="mask",
            image_column="image",
            transform=training_transform,
        ),
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        TiledSegmentationDataset(
            validation_dataframe,
            image_slicers,
            channels=channels,
            mask_column="mask",
            image_column="image",
            transform=validation_transform,
        ),
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    return train_loader, val_loader


def create_segmentation_training_dataframes_and_dataloaders(
    image_directories,
    mask_directories,
    save_dir,
    channels,
    validation_set_ratio=0.25,
    test_set_ratio=0.1,
    batch_size=5,
    num_workers=32,
    pin_memory=True,
    train_on_tiles=True,
    tiler_params=None,
    training_transform=None,
    validation_transform=None,
):
    (
        training_dataframe,
        validation_dataframe,
    ) = create_segmentation_training_dataframes(
        image_directories,
        mask_directories,
        save_dir,
        validation_set_ratio=validation_set_ratio,
        test_set_ratio=test_set_ratio,
    )
    train_loader, val_loader = create_segmentation_dataloaders(
        training_dataframe,
        validation_dataframe,
        channels,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=pin_memory,
        train_on_tiles=train_on_tiles,
        tiler_params=tiler_params,
        training_transform=training_transform,
        validation_transform=validation_transform,
    )
    return training_dataframe, validation_dataframe, train_loader, val_loader


def create_segmentation_dataloaders_from_filemap(
    filemap_path,
    save_dir,
    channels,
    image_column="image",
    mask_column="mask",
    validation_set_ratio=0.25,
    test_set_ratio=0.1,
    batch_size=5,
    num_workers=32,
    pin_memory=True,
    train_on_tiles=True,
    tiler_params=None,
    training_transform=None,
    validation_transform=None,
):
    dataframe = pd.read_csv(filemap_path)
    # rename image column to "image" and mask column to "mask"
    dataframe = dataframe.rename(columns={image_column: "image", mask_column: "mask"})

    training_dataframe, validation_dataframe, test_dataframe = split_dataset(
        filemap_path, validation_set_ratio, test_set_ratio
    )

    # backup the training and validation dataframes
    database_backup_dir = os.path.join(save_dir, "database_backup")
    current_date = datetime.datetime.now().strftime("%Y%m%d")
    os.makedirs(database_backup_dir, exist_ok=True)

    training_dataframe.to_csv(
        os.path.join(database_backup_dir, f"training_dataframe_{current_date}.csv"),
        index=False,
    )
    validation_dataframe.to_csv(
        os.path.join(database_backup_dir, f"validation_dataframe_{current_date}.csv"),
        index=False,
    )
    test_dataframe.to_csv(
        os.path.join(database_backup_dir, f"test_dataframe_{current_date}.csv"),
        index=False,
    )

    train_loader, val_loader = create_segmentation_dataloaders(
        training_dataframe,
        validation_dataframe,
        channels,
        batch_size=batch_size,
        num_workers=num_workers,
        pin_memory=pin_memory,
        train_on_tiles=train_on_tiles,
        tiler_params=tiler_params,
        training_transform=training_transform,
        validation_transform=validation_transform,
    )
    return training_dataframe, validation_dataframe, train_loader, val_loader


def create_classification_training_dataframes(
    ground_truth_csv_paths,
    image_columns,
    class_columns,
    save_dir,
    validation_set_ratio=0.25,
    test_set_ratio=0.1,
):
    if not isinstance(ground_truth_csv_paths, list):
        ground_truth_csv_paths = [ground_truth_csv_paths]
    if isinstance(image_columns, str):
        image_columns = [image_columns]
    if isinstance(class_columns, str):
        class_columns = [class_columns]

    if len(image_columns) == 1:
        image_columns = image_columns * len(ground_truth_csv_paths)
    if len(class_columns) == 1:
        image_columns = image_columns * len(class_columns)

    ground_truth_df = pd.DataFrame()
    for i, ground_truth_csv in enumerate(ground_truth_csv_paths):
        gt_df = pd.read_csv(ground_truth_csv)
        images = gt_df[image_columns[i]].values.tolist()
        classes = gt_df[class_columns[i]].values.tolist()
        new_gt_df = pd.DataFrame({"image": images, "class": classes})
        ground_truth_df = pd.concat([ground_truth_df, new_gt_df], ignore_index=True)

    training_dataframe, validation_dataframe, test_dataframe = split_dataset(
        ground_truth_df, validation_set_ratio, test_set_ratio
    )

    # backup the training and validation dataframes
    database_backup_dir = os.path.join(save_dir, "database_backup")
    current_date = datetime.datetime.now().strftime("%Y%m%d")
    os.makedirs(database_backup_dir, exist_ok=True)

    training_dataframe.to_csv(
        os.path.join(database_backup_dir, f"training_dataframe_{current_date}.csv"),
        index=False,
    )
    validation_dataframe.to_csv(
        os.path.join(database_backup_dir, f"validation_dataframe_{current_date}.csv"),
        index=False,
    )
    test_dataframe.to_csv(
        os.path.join(database_backup_dir, f"test_dataframe_{current_date}.csv"),
        index=False,
    )

    return training_dataframe, validation_dataframe


def create_classification_dataloaders(
    training_dataframe,
    validation_dataframe,
    channels,
    n_classes,
    batch_size=64,
    num_workers=32,
    pin_memory=True,
    training_transform=None,
    validation_transform=None,
):
    train_loader = DataLoader(
        ClassificationDataset(
            training_dataframe,
            channels=channels,
            n_classes=n_classes,
            class_column="class",
            image_column="image",
            transform=training_transform,
        ),
        batch_size=batch_size,
        shuffle=True,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    val_loader = DataLoader(
        ClassificationDataset(
            validation_dataframe,
            channels=channels,
            n_classes=n_classes,
            class_column="class",
            image_column="image",
            transform=validation_transform,
        ),
        batch_size=batch_size,
        shuffle=False,
        num_workers=num_workers,
        pin_memory=pin_memory,
    )
    return train_loader, val_loader



import numpy as np
from scipy.signal import medfilt
from scipy.signal import savgol_filter

from towbintools.data_analysis.time_series import correct_series_with_classification
from towbintools.data_analysis.time_series import smooth_series
from towbintools.data_analysis.time_series import smooth_series_classified
from towbintools.foundation.utils import interpolate_nans_infs


def compute_growth_rate_linear(
    series,
    time,
    ignore_start_fraction=0.0,
    ignore_end_fraction=0.0,
    savgol_filter_window=5,
    savgol_filter_order=3,
):
    """
    Compute the growth rate of a time series using linear regression.

    Parameters:
        series (np.ndarray): The series of values to compute the growth rate of.
        time (np.ndarray): The time data corresponding to the series.
        ignore_start_fraction (float): The fraction of points to ignore at the beginning of the series.
        ignore_end_fraction (float): The fraction of points to ignore at the end of the series.
        savgol_filter_window (int): The window size of the Savitzky-Golay filter.
        savgol_filter_order (int): The order of the Savitzky-Golay filter.

    Returns:
        float: The linear growth rate of the time series.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."

    # Compute the number of points to ignore at the beginning and end
    num_points = len(series)
    num_ignore_start = int(ignore_start_fraction * num_points)
    num_ignore_end = int(ignore_end_fraction * num_points)

    # Assert that the fraction of points to ignore is not too large
    assert (
        num_ignore_start + num_ignore_end < num_points
    ), "The fraction of points to ignore is too large."

    # Correctly handle slicing when ignore fractions are 0
    if num_ignore_end == 0:
        time = time[num_ignore_start:]
        series = series[num_ignore_start:]
    else:
        time = time[num_ignore_start:-num_ignore_end]
        series = series[num_ignore_start:-num_ignore_end]

    # Remove extreme outliers with a small median filter
    series = medfilt(series, 3)

    # Interpolate NaN and inf values
    series = interpolate_nans_infs(series)

    # Smooth the series time series a bit more with a Savitzky-Golay filter
    series = savgol_filter(series, savgol_filter_window, savgol_filter_order)

    # Compute linear regression
    slope, intercept = np.polyfit(time, series, 1)

    return slope


def compute_growth_rate_exponential(
    series,
    time,
    ignore_start_fraction=0.0,
    ignore_end_fraction=0.0,
    savgol_filter_window=5,
    savgol_filter_order=3,
):
    """
    Compute the growth rate of a time series using exponential regression.

    Parameters:
        series (np.ndarray): The series of values to compute the growth rate of.
        time (np.ndarray): The time data corresponding to the series.
        ignore_start_fraction (float): The fraction of points to ignore at the beginning of the series.
        ignore_end_fraction (float): The fraction of points to ignore at the end of the series.
        savgol_filter_window (int): The window size of the Savitzky-Golay filter.
        savgol_filter_order (int): The order of the Savitzky-Golay filter.

    Returns:
        float: The exponential growth rate of the time series.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."

    # Compute the number of points to ignore at the beginning and end
    num_points = len(series)
    num_ignore_start = int(ignore_start_fraction * num_points)
    num_ignore_end = int(ignore_end_fraction * num_points)

    # Assert that the fraction of points to ignore is not too large
    assert (
        num_ignore_start + num_ignore_end < num_points
    ), "The fraction of points to ignore is too large."

    # Correctly handle slicing when ignore fractions are 0
    if num_ignore_end == 0:
        time = time[num_ignore_start:]
        series = series[num_ignore_start:]
    else:
        time = time[num_ignore_start:-num_ignore_end]
        series = series[num_ignore_start:-num_ignore_end]

    # Remove extreme outliers with a small median filter
    series = medfilt(series, 3)

    # Interpolate NaN and inf values
    series = interpolate_nans_infs(series)

    # Smooth the series time series a bit more with a Savitzky-Golay filter
    series = savgol_filter(series, savgol_filter_window, savgol_filter_order)

    # Compute exponential regression
    slope, intercept = np.polyfit(time, np.log(series), 1)

    return slope


def compute_growth_rate_classified(
    series,
    time,
    worm_type,
    method="exponential",
    ignore_start_fraction=0.0,
    ignore_end_fraction=0.0,
    savgol_filter_window=5,
    savgol_filter_order=3,
):
    """
    Compute the growth rate of a time series after correcting the non-worm points by removing them and interpolating them back.

    Parameters:
        series (np.ndarray): The time series of values.
        time (np.ndarray): The time data corresponding to the series.
        worm_type (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.
        method (str): The method to use for computing the growth rate. Can be either 'exponential' or 'linear'.
        ignore_start_fraction (float): The fraction of points to ignore at the beginning of the series.
        ignore_end_fraction (float): The fraction of points to ignore at the end of the series.
        savgol_filter_window (int): The window size of the Savitzky-Golay filter.
        savgol_filter_order (int): The order of the Savitzky-Golay filter.

    Returns:
        float: The growth rate of the time series.
    """

    # Assert that the series, time, and worm_type have the same length
    assert (
        len(series) == len(time) == len(worm_type)
    ), "The series, time, and worm_type must have the same length."

    # Correct the series time series
    series_worms = correct_series_with_classification(series, worm_type)

    if method == "exponential":
        growth_rate = compute_growth_rate_exponential(
            series_worms,
            time,
            ignore_start_fraction,
            ignore_end_fraction,
            savgol_filter_window,
            savgol_filter_order,
        )
    elif method == "linear":
        growth_rate = compute_growth_rate_linear(
            series_worms,
            time,
            ignore_start_fraction,
            ignore_end_fraction,
            savgol_filter_window,
            savgol_filter_order,
        )

    return growth_rate


def compute_instantaneous_growth_rate(
    series,
    time,
    lmbda=0.0075,
    order=2,
    medfilt_window=5,
):
    """
    Compute the instantaneous growth rate of a time series.

    Parameters:
        series (np.ndarray): The series of values to compute the growth rate of.
        time (np.ndarray): The time data corresponding to the series.
        lmbda (float, optional): The smoothing parameter for the Whittaker-Eilers smoothing. Default provides good results for our volume curves when series_time is in hours. (default: 0.0075)
        order (int, optional): The order of the penalty of the Whittaker-Eilers smoother. (default: 2)
        medfilt_window (int, optional): The window size for the median filter. (default: 5)

    Returns:
        np.ndarray: The instantaneous growth rate of the time series.
    """

    # Assert that the series and time have the same length
    assert len(series) == len(time), "The series and time must have the same length."

    series = smooth_series(series, time, lmbda, order, medfilt_window)

    # Compute the instantaneous growth rate
    growth_rate = np.gradient(series, time)

    return growth_rate


def compute_instantaneous_growth_rate_classified(
    series,
    time,
    worm_type,
    lmbda=0.0075,
    order=2,
    medfilt_window=5,
):
    """
    Compute the instantaneous growth rate of a time series after correcting the non-worm points by removing them and interpolating them back.

    Parameters:
        series (np.ndarray): The time series of values.
        time (np.ndarray): The time data corresponding to the series.
        worm_type (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.
        lmbda (float, optional): The smoothing parameter for the Whittaker-Eilers smoothing. Default provides good results for our volume curves when series_time is in hours. (default: 0.0075)
        order (int, optional): The order of the penalty of the Whittaker-Eilers smoother. (default: 2)
        medfilt_window (int, optional): The window size for the median filter. (default: 5)

    Returns:
        np.ndarray: The instantaneous growth rate of the time series.
    """

    # Assert that the series, time, and worm_type have the same length
    assert (
        len(series) == len(time) == len(worm_type)
    ), "The series, time, and worm_type must have the same length."

    # Correct the series time series
    series = smooth_series_classified(
        series, time, worm_type, lmbda, order, medfilt_window
    )
    # Compute the instantaneous growth rate
    growth_rate = np.gradient(series, time)

    return growth_rate


def compute_growth_rate_per_larval_stage(
    series,
    time,
    worm_type,
    ecdysis,
    method="exponential",
    ignore_start_fraction=0.0,
    ignore_end_fraction=0.0,
    savgol_filter_window=5,
    savgol_filter_order=3,
):
    """
    Compute the growth rate of a time series per larval stage.

    Parameters:
        series (np.ndarray): The time series of values.
        time (np.ndarray): The time data corresponding to the series.
        worm_type (np.ndarray): The classification of the points as either 'worm' or 'egg' or 'error'.
        ecdysis (dict): The ecdysis events of the worm.
        method (str): The method to use for computing the growth rate. Can be either 'exponential' or 'linear'.
        ignore_start_fraction (float): The fraction of points to ignore at the beginning of the series.
        ignore_end_fraction (float): The fraction of points to ignore at the end of the series.
        savgol_filter_window (int): The window size of the Savitzky-Golay filter.
        savgol_filter_order (int): The order of the Savitzky-Golay filter.

    Returns:
        dict: The growth rate per larval stage.
    """

    # Assert that the series, time, and worm_type have the same length
    assert (
        len(series) == len(time) == len(worm_type)
    ), "The series, time, and worm_type, must have the same length."

    # Correct the series time series
    series_worms = correct_series_with_classification(series, worm_type)

    # extract ecdisis indices
    hatch_time = ecdysis["HatchTime"]
    M1 = ecdysis["M1"]
    M2 = ecdysis["M2"]
    M3 = ecdysis["M3"]
    M4 = ecdysis["M4"]

    growth_rates = {}

    # Compute the growth rate per larval stage
    for i, (start, end) in enumerate(zip([hatch_time, M1, M2, M3], [M1, M2, M3, M4])):
        # check if start or end is NaN
        if np.isnan(start) or np.isnan(end):
            growth_rates[f"L{i+1}"] = np.nan

        else:
            series_worms_stage = series_worms[start:end]
            time_stage = time[start:end]
            worm_type_stage = worm_type[start:end]

            growth_rate_stage = compute_growth_rate_classified(
                series_worms_stage,
                time_stage,
                worm_type_stage,
                method,
                ignore_start_fraction,
                ignore_end_fraction,
                savgol_filter_window,
                savgol_filter_order,
            )
            growth_rates[f"L{i+1}"] = growth_rate_stage

    return growth_rates


def compute_larval_stage_duration(ecdysis):
    """
    Compute the duration of each larval stage.

    Parameters:
        ecdysis (dict): The ecdysis events of the worm.

    Returns:
        dict: The duration of each larval stage.
    """

    # extract ecdisis indices
    hatch_time = ecdysis["HatchTime"]
    M1 = ecdysis["M1"]
    M2 = ecdysis["M2"]
    M3 = ecdysis["M3"]
    M4 = ecdysis["M4"]

    ls_durations = {}

    for i, (start, end) in enumerate(zip([hatch_time, M1, M2, M3], [M1, M2, M3, M4])):
        # check if start or end is NaN
        if np.isnan(start) or np.isnan(end):
            ls_durations[f"L{i+1}"] = np.nan
        else:
            ls_durations[f"L{i+1}"] = end - start

    return ls_durations



import numpy as np


def compute_fluorescence_in_mask(image, mask, aggregation="sum"):
    """Quantify fluorescence of an image in a mask.

    Parameters:
        image (np.ndarray): The image as a NumPy array.
        mask (np.ndarray): The binary mask as a NumPy array.
        aggregation (str): The aggregation method to use to quantify the fluorescence.
                             Can be one of 'sum', 'mean', 'median', 'max', 'min', or 'std'.

    Returns:
        float: Quantification of the fluorescence of the image in the mask.
    """

    if aggregation == "sum":
        return np.sum(image[mask > 0])
    elif aggregation == "mean":
        return np.mean(image[mask > 0])
    elif aggregation == "median":
        return np.median(image[mask > 0])
    elif aggregation == "max":
        return np.max(image[mask > 0])
    elif aggregation == "min":
        return np.min(image[mask > 0])
    elif aggregation == "std":
        return np.std(image[mask > 0])
    else:
        raise ValueError(
            'Aggregation must be one of "sum", "mean", "median", "max", "min", or "std".'
        )


def compute_background_fluorescence(image, foreground_mask, aggregation="mean"):
    """Estimate the background value of an image.

    Parameters:
        image (np.ndarray): The image as a NumPy array.
        foreground_mask (np.ndarray): The binary mask of the foreground as a NumPy array.
        aggregation (str): The aggregation method to use to estimate the background value.
                           Can be either 'mean', 'median', or 'min'.

    Returns:
        float: The estimated background value of the image.
    """

    background_mask = np.logical_not(foreground_mask > 0)

    if aggregation == "mean":
        return np.mean(image[background_mask])
    elif aggregation == "median":
        return np.median(image[background_mask])
    elif aggregation == "min":
        return np.min(image[background_mask])
    else:
        raise ValueError('Aggregation must be one of "mean", "median", or "min".')



from .growth_rate import compute_growth_rate_classified
from .growth_rate import compute_growth_rate_exponential
from .growth_rate import compute_growth_rate_linear
from .growth_rate import compute_growth_rate_per_larval_stage
from .growth_rate import compute_instantaneous_growth_rate
from .growth_rate import compute_instantaneous_growth_rate_classified
from .growth_rate import compute_larval_stage_duration
from .time_series import aggregate_interpolated_series
from .time_series import compute_exponential_series_at_time_classified
from .time_series import compute_series_at_time_classified
from .time_series import correct_series_with_classification
from .time_series import filter_series_with_classification
from .time_series import interpolate_entire_development
from .time_series import interpolate_entire_development_classified
from .time_series import rescale_and_aggregate
from .time_series import rescale_series
from .time_series import smooth_series_classified

__all__ = [
    "compute_growth_rate_classified",
    "compute_growth_rate_exponential",
    "compute_growth_rate_linear",
    "compute_growth_rate_per_larval_stage",
    "compute_instantaneous_growth_rate",
    "compute_instantaneous_growth_rate_classified",
    "compute_larval_stage_duration",
    "aggregate_interpolated_series",
    "compute_exponential_series_at_time_classified",
    "compute_series_at_time_classified",
    "correct_series_with_classification",
    "filter_series_with_classification",
    "interpolate_entire_development",
    "interpolate_entire_development_classified",
    "rescale_and_aggregate",
    "rescale_series",
    "smooth_series_classified",
]
